# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
import shutil
import stat
import tarfile
import tempfile
from pathlib import Path

import pmb.config
import pmb.helpers.apk
import pmb.helpers.http
import pmb.helpers.repo
import pmb.helpers.run
import pmb.parse.apkindex
from pmb.core.arch import Arch
from pmb.core.context import get_context
from pmb.helpers import logging
from pmb.helpers.exceptions import NonBugError


def read_signature_info(tar: tarfile.TarFile) -> tuple[str, str]:
    """
    Find various information about the signature that was used to sign
    /sbin/apk.static inside the archive (not to be confused with the normal apk
    archive signature!)

    :returns: (sigfilename, sigkey_path)
    """
    # Get signature filename and key
    prefix = "sbin/apk.static.SIGN.RSA.sha256."
    sigfilename = None
    for filename in tar.getnames():
        if filename.startswith(prefix):
            sigfilename = filename
            break
    if not sigfilename:
        raise RuntimeError(
            "Could not find signature filename in apk."
            " This means that your apk file is damaged."
            " Delete it and try again."
            " If the problem persists, fill out a bug report."
        )
    sigkey = sigfilename[len(prefix) :]
    logging.debug(f"sigfilename: {sigfilename}")
    logging.debug(f"sigkey: {sigkey}")

    # Get path to keyfile on disk
    sigkey_path = f"{pmb.config.apk_keys_path}/{sigkey}"
    if "/" in sigkey or not os.path.exists(sigkey_path):
        logging.debug(f"sigkey_path: {sigkey_path}")
        raise RuntimeError(f"Invalid signature key: {sigkey}")

    return (sigfilename, sigkey_path)


def extract_temp(tar: tarfile.TarFile, sigfilename: str) -> dict[str, dict]:
    """Extract apk.static and signature as temporary files."""
    ret = {
        "apk": {"filename": "sbin/apk.static", "temp_path": None},
        "sig": {"filename": sigfilename, "temp_path": None},
    }
    for ftype in ret:
        filename = ret[ftype]["filename"]
        if filename is None:
            raise AssertionError
        member = tar.getmember(filename)

        fd, path = tempfile.mkstemp(ftype, "pmbootstrap")
        with open(fd, "wb") as handle:
            ret[ftype]["temp_path"] = path
            extracted_file = tar.extractfile(member)
            if extracted_file is None:
                raise AssertionError
            shutil.copyfileobj(extracted_file, handle)

            logging.debug(f"extracted: {path}")
    return ret


def verify_signature(files: dict[str, dict], sigkey_path: str) -> None:
    """
    Verify the signature with openssl.

    :param files: return value from extract_temp()
    :raises RuntimeError: when verification failed and  removes temp files
    """
    logging.debug(f"Verify apk.static signature with {sigkey_path}")
    try:
        pmb.helpers.run.user(
            [
                "openssl",
                "dgst",
                "-sha256",
                "-verify",
                sigkey_path,
                "-signature",
                files["sig"]["temp_path"],
                files["apk"]["temp_path"],
            ]
        )
    except BaseException as exception:
        os.unlink(files["sig"]["temp_path"])
        os.unlink(files["apk"]["temp_path"])
        raise RuntimeError(
            "Failed to validate signature of apk.static."
            " Either openssl is not installed, or the"
            " download failed. Run 'pmbootstrap zap -hc' to"
            " delete the download and try again."
        ) from exception


def extract(version: str, apk_path: Path) -> None:
    """
    Extract everything to temporary locations, verify signatures and reported
    versions. When everything is right, move the extracted apk.static to the
    final location.
    """
    # Extract to a temporary path
    with tarfile.open(apk_path, "r:gz") as tar:
        sigfilename, sigkey_path = read_signature_info(tar)
        files = extract_temp(tar, sigfilename)

    # Verify signature
    verify_signature(files, sigkey_path)
    os.unlink(files["sig"]["temp_path"])
    temp_path = files["apk"]["temp_path"]

    # Verify the version that the extracted binary reports
    logging.debug(
        "Verify the version reported by the apk.static binary"
        f" (must match the package version {version})"
    )
    os.chmod(temp_path, os.stat(temp_path).st_mode | stat.S_IEXEC)
    version_bin = pmb.helpers.run.user_output([temp_path, "--version"])
    version_bin = version_bin.split(" ")[1].split(",")[0]
    if not version.startswith(f"{version_bin}"):
        os.unlink(temp_path)
        raise RuntimeError(
            f"Downloaded apk-tools-static-{version}.apk, but the apk binary inside that package"
            f" reports to be version: {version_bin}! This could potentially be caused by a"
            " downgrade attack from a malicious server or a bug in pmbootstrap. You can try"
            " switching servers, see https://docs.postmarketos.org/pmbootstrap/mirrors.html"
        )

    # Move it to the right path
    target_path = get_context().config.work / "apk.static"
    shutil.move(temp_path, target_path)


def download(file: str) -> Path:
    """Download a single file from an Alpine mirror."""
    channel_cfg = pmb.config.pmaports.read_config_channel()
    mirrordir = channel_cfg["mirrordir_alpine"]
    base_url = f"{get_context().config.mirrors['alpine']}{mirrordir}/main/{Arch.native()}"
    return pmb.helpers.http.download(f"{base_url}/{file}", file)


def init() -> None:
    """Download, verify, extract $WORK/apk.static."""
    # Get and parse the APKINDEX. alpine_apkindex_path() will implicitly
    # download the APKINDEX file if it's missing.
    apkindex = pmb.helpers.repo.alpine_apkindex_path("main")
    index_data = pmb.parse.apkindex.package("apk-tools-static", indexes=[apkindex])

    if index_data is None:
        raise RuntimeError("Could not find apk-tools-static in APKINDEX!")

    version = index_data.version

    # Verify the apk-tools-static version
    try:
        pmb.helpers.apk.check_outdated(version)
    except RuntimeError as exception:
        raise NonBugError(f"{exception}. Run 'pmbootstrap update', then try again.") from exception

    # Download, extract, verify apk-tools-static
    apk_name = f"apk-tools-static-{version}.apk"
    apk_static = download(apk_name)
    extract(version, apk_static)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
import re
from pathlib import Path
from typing import Any

import pmb.chroot
import pmb.config
import pmb.helpers.pmaports
import pmb.helpers.run
from pmb.core.context import get_context
from pmb.core.pkgrepo import pkgrepo_default_path
from pmb.helpers import logging
from pmb.helpers.exceptions import NonBugError


def folder_size(path: Path) -> int:
    """
    Run `du` to calculate the size of a folder.

    (this is less code and faster than doing the same task in pure Python)
    This result is only approximately right, but good enough for pmbootstrap's use case (#760).

    :returns: folder size in kilobytes
    """
    output = pmb.helpers.run.root(["du", "-ks", path], output_return=True)

    # Only look at last line to filter out sudo garbage (#1766)
    last_line = output.split("\n")[-2]

    return int(last_line.split("\t")[0])


def check_grsec() -> None:
    """
    Check if the current kernel is based on the grsec patchset.

    Also check if the chroot_deny_chmod option is enabled.
    Raise an exception in that case, with a link to the issue. Otherwise, do nothing.
    """
    path = "/proc/sys/kernel/grsecurity/chroot_deny_chmod"
    if not os.path.exists(path):
        return

    raise RuntimeError(
        "You're running a kernel based on the grsec patchset. This is not supported."
    )


def check_binfmt_misc() -> None:
    """
    Check if the 'binfmt_misc' module is loaded.

    This is done by checking, if /proc/sys/fs/binfmt_misc/ exists.
    If it exists, then do nothing.
    Otherwise, load the module and mount binfmt_misc.
    If that fails as well, raise an exception pointing the user to the wiki.
    """
    path = "/proc/sys/fs/binfmt_misc/status"
    if os.path.exists(path):
        return

    # check=False: this might be built-in instead of being a module
    pmb.helpers.run.root(["modprobe", "binfmt_misc"], check=False)

    # check=False: we check it below and print a more helpful message on error
    pmb.helpers.run.root(
        ["mount", "-t", "binfmt_misc", "none", "/proc/sys/fs/binfmt_misc"], check=False
    )

    if not os.path.exists(path):
        link = "https://postmarketos.org/binfmt_misc"
        raise RuntimeError(f"Failed to set up binfmt_misc, see: {link}")


def migrate_success(work: Path, version: int) -> None:
    logging.info("Migration to version " + str(version) + " done")
    work.joinpath("version").write_text(str(version) + "\n")


def migrate_work_folder() -> None:
    # Read current version
    context = get_context()
    current = 0
    suffix: str | None = None
    current_with_suffix = ""
    path = context.config.work / "version"
    if os.path.exists(path):
        # pmb 2.3.x added a suffix due to conflicting work versions
        # We need to be able to handle that going forward
        current_with_suffix = path.read_text().rstrip()
        version_parts = current_with_suffix.split("-")
        current = int(version_parts[0])
        if len(version_parts) == 2:
            suffix = version_parts[1]

    # Compare version, print warning or do nothing
    required = pmb.config.work_version
    if current == required:
        return
    logging.info(
        "WARNING: Your work folder version needs to be migrated"
        f" (from version {current_with_suffix} to {required})!"
    )

    # version 6 and version 7 from 2.3.x branch are equivalent for this and we need to migrate
    if current == 6 or (current == 7 and suffix == "2.x"):
        # Ask for confirmation
        logging.info("Changelog:")
        logging.info("* Major refactor for pmb 3.0.0")
        logging.info("Migration will do the following:")
        logging.info("* Zap your chroots")
        if not pmb.helpers.cli.confirm():
            raise RuntimeError("Aborted.")

        # Zap chroots
        pmb.chroot.zap(False)

        # Update version file
        if suffix == "2.x":
            # If we come from 7-2.x, then we already updated the git urls and
            # can skip the 7->8 migration step
            migrate_success(context.config.work, 8)
            current = 8
        else:
            migrate_success(context.config.work, 7)
            current = 7

    if current == 7:
        # Ask for confirmation
        logging.info("Changelog:")
        logging.info("* Moved from gitlab.com to gitlab.postmarketOS.org")
        logging.info("Migration will do the following:")
        logging.info("* Update your pmaports remote URL")
        if not pmb.helpers.cli.confirm():
            raise RuntimeError("Aborted.")

        pmb.helpers.git.migrate_upstream_remote()
        try:
            pmb.helpers.git.get_upstream_remote(pkgrepo_default_path())
        except RuntimeError as exception:
            logging.error(
                "Couldn't find new upstream remote, migration failed."
                " Please try updating the remote manually with:\n"
                f" $ git -C '{pkgrepo_default_path()}' remote set-url origin 'https://gitlab.postmarketos.org/postmarketOS/pmaports.git'"
            )
            raise RuntimeError("Migration failed.") from exception

        # Update version file
        migrate_success(context.config.work, 8)
        current = 8

    # Can't migrate, user must delete it
    if current != required:
        raise NonBugError(
            "Sorry, we can't migrate that automatically. Please"
            " run 'pmbootstrap shutdown', then delete your"
            " current work folder manually ('sudo rm -rf "
            f"{context.config.work}') and start over with 'pmbootstrap"
            " init'. All your binary packages and caches will"
            " be lost."
        )


def normalize_hostname(hostname: str) -> str:
    """
    Fixup default hostnames so that they don't fail validate_hostname()

    This should not be called on user-chosen hostnames as those should fail
    """
    # Truncate length
    if len(hostname) > 63:
        hostname = hostname[:63]

    # Replace underscores with dashes
    if "_" in hostname:
        hostname = hostname.replace("_", "-")

    # We shouldn't have to fix the rest of the regex because the APKBUILDs'
    # device names shouldn't have any more invalid characters

    return hostname


def validate_hostname(hostname: str) -> bool:
    """
    Check whether the string is a valid hostname.

    Check is performed according to
    <http://en.wikipedia.org/wiki/Hostname#Restrictions_on_valid_host_names>
    """
    # Check length
    if len(hostname) > 63:
        logging.fatal(f"ERROR: Hostname '{hostname}' is too long.")
        return False

    if len(hostname) == 0:
        logging.fatal("ERROR: Hostnames must be at least one character.")
        return False

    # Check that it only contains valid chars
    if not re.match(r"^[0-9a-z-\.]*$", hostname):
        logging.fatal(
            "ERROR: Hostname must only contain letters (a-z), digits (0-9), minus signs "
            "(-), or periods (.)"
        )
        return False

    # Check that doesn't begin or end with a minus sign or period
    if re.search(r"^-|^\.|-$|\.$", hostname):
        logging.fatal("ERROR: Hostname must not begin or end with a minus sign or period")
        return False

    return True


"""
pmbootstrap uses this dictionary to save the result of expensive
results, so they work a lot faster the next time they are needed in the
same session. Usually the cache is written to and read from in the same
Python file, with code similar to the following:

def lookup(key):
    if key in pmb.helpers.other.cache["mycache"]:
        return pmb.helpers.other.cache["mycache"][key]
    ret = expensive_operation(args, key)
    pmb.helpers.other.cache["mycache"][key] = ret
    return ret
"""
cache: dict[str, Any] = {
    "apkindex": {},
}



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import datetime
import functools
import operator
from collections.abc import Callable
from pathlib import Path
from typing import Any, TypedDict

import pmb.build
import pmb.build.autodetect
import pmb.chroot
import pmb.chroot.apk
import pmb.config.pmaports
import pmb.helpers.package
import pmb.helpers.pmaports
import pmb.helpers.repo
import pmb.parse
import pmb.parse.apkindex
from pmb.build.other import BuildStatus
from pmb.core import Chroot
from pmb.core.arch import Arch
from pmb.core.context import Context, get_context
from pmb.core.pkgrepo import pkgrepo_relative_path
from pmb.helpers import logging
from pmb.helpers.exceptions import BuildFailedError, CommandFailedError, NonBugError
from pmb.types import Apkbuild, CrossCompile

from .backend import BootstrapStage, run_abuild


def check_build_for_arch(pkgname: str, arch: Arch) -> bool:
    """
    Check if pmaport can be built or exists as binary for a specific arch.

    :returns: * True when it can be built
              * False when it can't be built, but exists in a binary repo
                (e.g. temp/mesa can't be built for x86_64, but Alpine has it)
    :raises: RuntimeError if the package can't be built for the given arch and
             does not exist as binary package.
    """
    context = get_context()
    # Check for pmaport with arch
    if pmb.helpers.package.check_arch(pkgname, arch, False):
        return True

    # Check for binary package
    binary = pmb.parse.apkindex.package(pkgname, arch, False)
    if binary:
        pmaport = pmb.helpers.pmaports.get(pkgname)
        pmaport_version = pmaport["pkgver"] + "-r" + pmaport["pkgrel"]
        logging.debug(
            pkgname + ": found pmaport (" + pmaport_version + ") and"
            " binary package (" + binary.version + ", from"
            " postmarketOS or Alpine), but pmaport can't be built"
            f" for {arch} -> using binary package"
        )
        return False

    # No binary package exists and can't build it
    logging.info("NOTE: You can edit the 'arch=' line inside the APKBUILD")
    if context.command == "build":
        logging.info(
            "NOTE: Alternatively, use --arch to build for another"
            " architecture ('pmbootstrap build --arch=armhf " + pkgname + "')"
        )
    raise RuntimeError(f"Can't build '{pkgname}' for architecture {arch}")


def get_depends(context: Context, apkbuild: dict[str, Any]) -> list[str]:
    """
    Alpine's abuild always builds/installs the "depends" and "makedepends" of a package
    before building it.

    We used to only care about "makedepends"
    and it's still possible to ignore the depends with --ignore-depends.

    :returns: list of dependency pkgnames (eg. ["sdl2", "sdl2_net"])
    """
    # Read makedepends and depends
    if apkbuild["makedepends"]:
        ret = list(apkbuild["makedepends"])
    else:
        ret = list(apkbuild["makedepends_build"]) + list(apkbuild["makedepends_host"])
    if "!check" not in apkbuild["options"]:
        ret += apkbuild["checkdepends"]
    if not context.ignore_depends:
        ret += apkbuild["depends"]
    ret = sorted(set(ret))

    # Don't recurse forever when a package depends on itself (#948)
    for pkgname in [apkbuild["pkgname"], *apkbuild["subpackages"].keys()]:
        if pkgname in ret:
            logging.verbose(apkbuild["pkgname"] + ": ignoring dependency on itself: " + pkgname)
            ret.remove(pkgname)

    # FIXME: is this needed? is this sensible?
    return list(filter(lambda x: not x.startswith("!"), ret))


def get_pkgver(original_pkgver: str, original_source: bool = False) -> str:
    """
    Get the original pkgver when using the original source.

    Otherwise, get the pkgver with an appended suffix of current date and time.
    For example: ``_p20180218550502``
    When appending the suffix, an existing suffix (e.g. ``_git20171231``) gets
    replaced.

    :param original_pkgver: unmodified pkgver from the package's APKBUILD.
    :param original_source: the original source is used instead of overriding
                            it with --src.
    """
    if original_source:
        return original_pkgver

    # Append current date
    no_suffix = original_pkgver.split("_", 1)[0]
    now = datetime.datetime.now()
    new_suffix = "_p" + now.strftime("%Y%m%d%H%M%S")
    return no_suffix + new_suffix


def output_path(arch: Arch, pkgname: str, pkgver: str, pkgrel: str) -> Path:
    # Yeahp, you can just treat an Arch like a path!
    return arch / f"{pkgname}-{pkgver}-r{pkgrel}.apk"


def finish(
    apkbuild: dict[str, Any],
    channel: str,
    arch: Arch,
    output: Path,
    chroot: Chroot,
    strict: bool = False,
) -> None:
    """Various finishing tasks that need to be done after a build."""
    # Verify output file
    out_dir = get_context().config.work / "packages" / channel
    if not (out_dir / output).exists():
        raise RuntimeError(f"Package not found after build: {(out_dir / output)}")

    # Clear APKINDEX cache (we only parse APKINDEX files once per session and
    # cache the result for faster dependency resolving, but after we built a
    # package we need to parse it again)
    pmb.parse.apkindex.clear_cache(out_dir / arch / "APKINDEX.tar.gz")

    # Zap chroots (strict mode)
    zap_reason = None
    if "pmb:strict" in apkbuild["options"]:
        zap_reason = "pmb:strict in APKBUILD"
    elif strict:
        zap_reason = "running with --strict"

    if zap_reason:
        logging.info(f"({chroot}) zapping chroots ({zap_reason})")
        pmb.chroot.zap(False)

    logging.info(f"@YELLOW@=>@END@ @BLUE@{channel}/{apkbuild['pkgname']}@END@: Done!")

    # If we just built a package which is used to build other packages, then
    # update the buildroot to use the newly built version.
    if not zap_reason and apkbuild["pkgname"] in pmb.config.build_packages:
        logging.info(
            f"NOTE: Updating package {apkbuild['pkgname']} in buildroot since it's"
            " used for building..."
        )
        pmb.chroot.apk.install([apkbuild["pkgname"]], chroot, build=False, quiet=True)


_package_cache: dict[str, list[str]] = {}


def is_cached_or_cache(arch: Arch, pkgname: str) -> bool:
    """
    Check if a package is in the visited packages cache, if not
    then mark it as visited. We must mark as visited before building
    to break cyclical dependency loops.
    """
    key = str(arch)
    if key not in _package_cache:
        _package_cache[key] = []

    visited = pkgname in _package_cache[key]
    if not visited:
        _package_cache[key].append(pkgname)
    else:
        logging.debug(f"{key}/{pkgname}: marked for build")
    return visited


def get_apkbuild(pkgname: str) -> tuple[Path | None, Apkbuild | None]:
    """
    Parse the APKBUILD path for pkgname.

    When there is none, try to find it in the binary package APKINDEX files or raise an exception.

    :param pkgname: package name to be built, as specified in the APKBUILD
    :returns: None or parsed APKBUILD
    """
    # Get pmaport, skip upstream only packages
    pmaport, apkbuild = pmb.helpers.pmaports.get_with_path(pkgname, False)
    if pmaport:
        pmaport = pkgrepo_relative_path(pmaport)[0]
        return pmaport, apkbuild

    return None, None


class BuildQueueItem(TypedDict):
    name: str
    arch: Arch  # Arch to build for
    aports: str
    apkbuild: dict[str, Any]
    has_binary: bool  # A binary package exists (even if outdated)
    pkgver: str
    output_path: Path
    channel: str
    depends: list[str]
    cross: CrossCompile


def has_cyclical_dependency(
    unmet_deps: dict[str, list[str]], item: BuildQueueItem, dep: str
) -> bool:
    pkgnames = [item["name"], *item["apkbuild"]["subpackages"].keys()]

    return any(pkgname in unmet_deps.get(dep, []) for pkgname in pkgnames)


def prioritise_build_queue(disarray: list[BuildQueueItem]) -> list[BuildQueueItem]:
    """Figure out The Correct Order to build packages in, or bail."""
    queue: list[BuildQueueItem] = []
    # (name, unmet_dep) of all unmet dependencies
    unmet_deps: dict[str, list[str]] = {}

    # build our base build packages first. This relies on
    # the build_packages array being in the correct order!
    for pkgname in pmb.config.build_packages:
        for item in disarray:
            if item["name"] == pkgname:
                queue.append(item)
                disarray.remove(item)
                break

    # list of packages in pmaports
    all_pkgnames = []
    for item in disarray:
        all_pkgnames.append(item["name"])
        all_pkgnames += item["apkbuild"]["subpackages"].keys()

    def queue_item(item: BuildQueueItem) -> None:
        queue.append(item)
        disarray.remove(item)
        all_pkgnames.remove(item["name"])
        for subpkg in item["apkbuild"]["subpackages"]:
            all_pkgnames.remove(subpkg)

        unmet_deps.pop(item["name"], None)

    stuck = False
    while disarray and not stuck:
        stuck = True
        for item in disarray:
            if not item["depends"]:
                queue_item(item)
                stuck = False
                break

            # If a dependency hasn't been queued yet, skip until it has been
            missing_deps = False
            for dep in item["depends"]:
                # This might be a subpkgname, replace with the main pkgname
                # (e.g."linux-pam-dev" -> "linux-pam")
                dep_data = pmb.helpers.package.get(
                    dep, item["arch"], must_exist=False, try_other_arches=False
                )
                if not dep_data:
                    raise NonBugError(f"{item['name']}: dependency not found: {dep}")
                dep = dep_data.pkgname

                # If the dependency is a subpackage we can safely ignore it
                if dep in item["apkbuild"]["subpackages"]:
                    continue

                if dep in all_pkgnames and dep_data.from_pmaports:
                    unmet_deps.setdefault(item["name"], []).append(dep)
                    missing_deps = True

                    if has_cyclical_dependency(unmet_deps, item, dep):
                        # If a binary package exists for item, we can queue it
                        # safely and dep will be queued on a future iteration
                        if item["has_binary"]:
                            logging.warning(
                                f"WARNING: cyclical build dependency: building {item['name']} with binary package of {dep}"
                            )
                            queue_item(item)
                            stuck = False
                            break
                        else:
                            logging.warning(
                                f"WARNING: cyclical build dependency: can't build {item['name']}, no binary package for {dep}"
                            )
                    else:
                        logging.debug(
                            f"{item['name']}: missing dependency {dep}, trying to queue other packages first"
                        )

            if missing_deps:
                continue

            # We're probably good to go??
            queue_item(item)
            stuck = False
            break

    if stuck:
        logging.error("Remaining packages:")
        for unmet_dep in unmet_deps:
            logging.error(f"* {unmet_dep}")
        raise NonBugError("Can't resolve build order of remaining packages!")

    return queue


def process_package(
    context: Context,
    queue_build: Callable,
    pkgname: str,
    arch: Arch | None,
    fallback_arch: Arch,
    force: bool,
    from_src: bool,
) -> list[str]:
    """:param arch: Set if we should build for a specific arch."""
    # Only build when APKBUILD exists
    base_aports, base_apkbuild = get_apkbuild(pkgname)
    if not base_apkbuild:
        # We allow this function to be called for packages that aren't in pmaports
        # and just do nothing in this case. However this can be quite confusing
        # when building an Alpine package with --src since we'll just do nothing
        if pmb.parse.apkindex.providers(pkgname, fallback_arch, False):
            if from_src:
                raise NonBugError(
                    f"Package {pkgname} is not in pmaports, but exists in Alpine."
                    " to build it with --src you first need to fork it to pmaports."
                    f" Please run 'pmbootstrap aportgen --fork-alpine {pkgname}' and then"
                    " try again"
                )
            return []
        raise RuntimeError(f"{pkgname}: Could not find it in pmaports or any APKINDEX!")

    if arch is None:
        arch = pmb.build.autodetect.arch(base_apkbuild)

    if is_cached_or_cache(arch, pkgname) and not force:
        logging.verbose(f"S{arch}/{pkgname}: already queued")
        return []

    logging.debug(f"{arch}/{pkgname}: Generating dependency tree")
    # Add the package to the build queue
    base_depends = get_depends(context, base_apkbuild)

    depends = base_depends.copy()

    base_build_status = BuildStatus.NEW if force else BuildStatus.UNNECESSARY
    if not base_build_status.necessary():
        base_build_status = pmb.build.get_status(arch, base_apkbuild)
    if base_build_status.necessary() and not check_build_for_arch(pkgname, arch):
        base_build_status = BuildStatus.UNNECESSARY

    # FIXME: We descend into the package dependencies even if we aren't going to
    # build it because this is the only way we warn about outdated packages. When
    # you run pmbootstrap install you expect to be warned that you bumped some random
    # utility but forgot to build it.
    # This ought to be fixed, ideally by having the actual dependency parsing stuff
    # here all abstracted away, then the package building code just becomes a couple
    # of callback functions to determine if a package should be built and build it,
    # and we can have another feature to walk the dependency graph and determine if
    # you maybe forgot to build a package, since we only care about that during
    # installation.
    if base_build_status.necessary():
        queue_build(base_aports, base_apkbuild, base_depends)

    # Also traverse subpackage depends, they shouldn't be included in base_depends since they
    # aren't needed for building (and can conflict with depends for other subpackages)
    depends += functools.reduce(
        operator.iadd,
        (sp["depends"] if sp else [] for sp in base_apkbuild["subpackages"].values()),
        [],
    )

    parent = pkgname
    while len(depends):
        # FIXME: pop(0) is really quite slow!
        dep = depends.pop(0)
        if is_cached_or_cache(arch, pmb.helpers.package.remove_operators(dep)):
            continue

        aports, apkbuild = get_apkbuild(dep)
        if not apkbuild:
            continue

        cross = pmb.build.autodetect.crosscompile(apkbuild, arch)

        if context.no_depends:
            pmb.helpers.repo.update(arch)
            dep_arch = Arch.native() if cross == CrossCompile.CROSS_NATIVE2 else arch
            if not pmb.parse.apkindex.package(dep, dep_arch, False):
                raise RuntimeError(
                    "Missing binary package for dependency '" + dep + "' of '" + parent + "', but"
                    " pmbootstrap won't build any depends since"
                    " it was started with --no-depends."
                )

        bstatus = pmb.build.get_status(arch, apkbuild)
        if bstatus.necessary() and dep not in pmb.config.build_packages:
            if context.no_depends:
                raise RuntimeError(
                    f"Binary package for dependency '{dep}'"
                    f" of '{parent}' is outdated, but"
                    f" pmbootstrap won't build any depends"
                    f" since it was started with --no-depends."
                )

            deps = get_depends(context, apkbuild)
            logging.debug(
                f"BUILDQUEUE: queue {dep} (dependency of {parent}) for build, reason: {bstatus}"
            )
            queue_build(aports, apkbuild, deps, cross)

            subpkg_deps: list[str] = functools.reduce(
                operator.iadd,
                (sp["depends"] if sp else [] for sp in apkbuild["subpackages"].values()),
                [],
            )
            logging.verbose(
                f"{arch}/{dep}: Inserting {len(deps)} dependencies and {len(subpkg_deps)} from subpackages"
            )
            depends = subpkg_deps + deps + depends
            parent = dep

    return depends


def packages(
    context: Context,
    pkgnames: list[str],
    arch: Arch | None = None,
    force: bool = False,
    strict: bool = False,
    src: str | None = None,
    bootstrap_stage: int = BootstrapStage.NONE,
    log_callback: Callable | None = None,
) -> list[str]:
    """
    Build a package and its dependencies with Alpine Linux' abuild.

    :param pkgname: package name to be built, as specified in the APKBUILD
    :param arch: architecture we're building for (default: native)
    :param force: always build, even if not necessary
    :param strict: avoid building with irrelevant dependencies installed by
                   letting abuild install and uninstall all dependencies.
    :param src: override source used to build the package with a local folder
    :param bootstrap_stage: pass a BOOTSTRAP= env var with the value to abuild
    :param log_callback: function to call before building each package instead of
                         logging. It should accept a single BuildQueueItem parameter.
    :returns: None if the build was not necessary
              output path relative to the packages folder ("armhf/ab-1-r2.apk")
    """
    global _package_cache

    build_queue: list[BuildQueueItem] = []
    built_packages: set[str] = set()

    # We want to build packages in the order they're given here. Due to how we
    # walk the package dependencies, reverse the list so that when we later
    # reverse the build queue we'll be back in the right order.
    pkgnames.reverse()

    # Add a package to the build queue, fetch it's dependency, and
    # add record build helpers to installed (e.g. sccache)
    def queue_build(
        aports: Path,
        apkbuild: dict[str, Any],
        depends: list[str],
        cross: CrossCompile | None = None,
    ) -> list[str]:
        # Skip if already queued
        name = apkbuild["pkgname"]
        if any(item["name"] == name for item in build_queue):
            return []

        pkg_arch = pmb.build.autodetect.arch(apkbuild) if arch is None else arch
        cross = cross or pmb.build.autodetect.crosscompile(apkbuild, pkg_arch)
        pkgver = get_pkgver(apkbuild["pkgver"], src is None)
        channel = pmb.config.pmaports.read_config(aports)["channel"]
        index_data = pmb.parse.apkindex.package(name, arch, False)
        # Make sure we aren't building a package that will never be used! This can happen if
        # building with --src with an outdated pmaports checkout. Unless --force is used
        # in which case we assume it was intentional.
        if (
            index_data
            and pmb.parse.version.compare(index_data.version, f"{pkgver}-r{apkbuild['pkgrel']}")
            == 1
        ):
            if force:
                logging.warning(
                    f"WARNING: A binary package for {name} has a newer version ({index_data.version})"
                    f" than the source ({pkgver}-{apkbuild['pkgrel']}). The package to be build will"
                    f" not be installed automatically, use 'apk add {name}={pkgver}-r{apkbuild['pkgrel']}'"
                    " to install it."
                )
            else:
                raise NonBugError(
                    f"A binary package for {name} has a newer version ({index_data.version})"
                    f" than the source ({pkgver}-{apkbuild['pkgrel']}). Please ensure your pmaports branch is up"
                    " to date and that you don't have a newer version of the package in your local"
                    f" binary repo ({context.config.work / 'packages' / channel / pkg_arch})."
                )
        build_queue.append(
            {
                "name": name,
                "arch": pkg_arch,
                "aports": aports.name,  # the pmaports source repo (e.g. "systemd")
                "apkbuild": apkbuild,
                "has_binary": bool(index_data),
                "pkgver": pkgver,
                "output_path": output_path(
                    pkg_arch, apkbuild["pkgname"], pkgver, apkbuild["pkgrel"]
                ),
                "channel": channel,
                "depends": depends,
                "cross": cross,
            }
        )

        # If we just queued a package that was request to be built explicitly then
        # record it, since we return which packages we actually built
        if apkbuild["pkgname"] in pkgnames:
            built_packages.add(apkbuild["pkgname"])

        return depends

    if src and len(pkgnames) > 1:
        raise RuntimeError("Can't build multiple packages with --src")

    logging.debug(f"Attempting to build: {', '.join(pkgnames)}")

    # We sorta-kind maybe supported building packages for multiple architectures in
    # a single called to packages(). We need to do a check to make sure that the user
    # didn't specify a package that doesn't exist, and we can't just check the source repo
    # since we might get called with some perhaps bogus packages that do exist in the binary
    # repo but not in the source one, but we need to error if we get a package that doesn't
    # exist anywhere, as something is clearly wrong for that to happen.
    # The problem is the APKINDEX parsing code doesn't have a way to check all architectures
    # so we need this hack.
    fallback_arch = arch if arch is not None else pmb.build.autodetect.arch(pkgnames[0])
    # Get existing binary package indexes
    pmb.helpers.repo.update(fallback_arch)

    # Process the packages we've been asked to build, queuing up any
    # dependencies that need building as well as the package itself
    all_dependencies: list[str] = []
    for pkgname in pkgnames:
        all_dependencies += process_package(
            context, queue_build, pkgname, arch, fallback_arch, force, src is not None
        )

    # If any of our common build packages need to be built and are missing, then add them
    # to the queue so they're built first. This is necessary so that our abuild fork is
    # always built first (for example). For now assume that if building in strict mode we
    # should skip this step, but we might want to revisit this later.
    if not src:
        for pkgname in pmb.config.build_packages:
            if pkgname not in pkgnames:
                aport, apkbuild = get_apkbuild(pkgname)
                if not aport or not apkbuild:
                    continue
                bstatus = pmb.build.get_status(arch, apkbuild)
                if bstatus.necessary():
                    if strict:
                        raise RuntimeError(
                            f"Strict mode enabled and build package {pkgname} needs building."
                            " Please build it manually first or build without --strict to build"
                            " it automatically."
                        )
                    queue_build(aport, apkbuild, get_depends(context, apkbuild))

    if not len(build_queue):
        return []

    build_queue = prioritise_build_queue(build_queue)

    qlen = len(build_queue)
    logging.info(f"Building @BLUE@{qlen}@END@ package{'s' if qlen > 1 else ''}")
    for item in build_queue:
        logging.info(f"   @BLUE@*@END@ {item['channel']}/{item['name']}")

    if len(build_queue) > 1 and src:
        raise RuntimeError(
            "Additional packages need building, please build them first and then"
            " build the package with --src again."
        )

    cross: CrossCompile | None = None
    prev_cross: CrossCompile | None = None
    hostchroot = None  # buildroot for the architecture we're building for

    total_pkgs = len(build_queue)
    for count, pkg in enumerate(build_queue, 1):
        prev_cross = cross
        cross = pkg["cross"]
        pkg_arch = pkg["arch"]
        hostchroot = cross.host_chroot(pkg_arch)
        buildchroot = cross.build_chroot(pkg_arch)
        apkbuild = pkg["apkbuild"]

        channel = pkg["channel"]
        output = pkg["output_path"]
        if not log_callback:
            logging.info(
                f"@YELLOW@=> ({count}/{total_pkgs})@END@ @BLUE@{channel}/{pkg['name']}@END@: Installing dependencies"
            )
        else:
            log_callback(pkg)

        # FIXME: this is only used to detect special compilers and a workaround for rust
        # in pmb.build.init_compiler(), this should all be refactored and enforce correct
        # APKBUILDs rather than trying to hack things in here
        pkg_depends = list(
            {
                *pkg["depends"],
                *apkbuild.get("makedepends", []),
                *apkbuild.get("makedepends_build", []),
                *apkbuild.get("makedepends_host", []),
            }
        )

        # One time chroot initialization
        if hostchroot != buildchroot:
            pmb.build.init(hostchroot)
        if pmb.build.init(buildchroot):
            pmb.build.other.configure_abuild(buildchroot)
            pmb.build.other.configure_ccache(buildchroot)
            if "rust" in all_dependencies or "cargo" in all_dependencies:
                pmb.chroot.apk.install(["sccache"], buildchroot)

        if (strict or cross != prev_cross) and cross.enabled():
            pmb.build.init_compiler(context, pkg_depends, cross, pkg_arch)
            if cross == CrossCompile.CROSSDIRECT:
                pmb.chroot.mount_native_into_foreign(buildchroot)

        depends_build: list[str] = []
        depends_host: list[str] = []
        # * makedepends_build are dependencies that should be installed in the native
        #   chroot (e.g. 'meson').
        # * makedepends_host are dependencies that should be
        #   in the chroot for the architecture we're building for (e.g. 'libevdev-dev').
        if apkbuild["makedepends_host"]:
            depends_host = apkbuild["makedepends_host"]
        else:
            depends_host = apkbuild["makedepends"]
        if apkbuild["makedepends_build"]:
            # If we have makedepends_build but not host then just subtract
            # the host depends from the main depends
            if "makedepends_host" not in apkbuild:
                depends_host = list(set(depends_host) - set(apkbuild["makedepends_build"]))
            depends_build = apkbuild["makedepends_build"]
        else:
            depends_build = apkbuild["makedepends"]
            if depends_build and depends_host and cross == CrossCompile.CROSS_NATIVE2:
                logging.warning(
                    "WARNING: makedepends not split into _host and _build variants."
                    " Trying to install all makedepends in both environments, please"
                    " adjust your APKBUILD to specify host dependencies (e.g. libevdev-dev)"
                    " and build dependencies (e.g. meson) separately."
                )
        if "!check" not in apkbuild["options"]:
            depends_build += apkbuild["checkdepends"]

        # If build with --src then we need rsync to copy in the source files.
        if src:
            depends_build.append("rsync")

        if hostchroot == buildchroot:
            pmb.chroot.apk.install(set(depends_host + depends_build), hostchroot, build=False)
        else:
            if depends_host:
                logging.info("*** Install host dependencies")
                pmb.chroot.apk.install(depends_host, hostchroot, build=False)
            if depends_build:
                logging.info("*** Install build dependencies")
                pmb.chroot.apk.install(depends_build, buildchroot, build=False)

        # Build and finish up
        msg = f"@YELLOW@=>@END@ @BLUE@{channel}/{pkg['name']}@END@: Building package"
        if cross != CrossCompile.UNNECESSARY:
            msg += f" (cross compiling: {cross})"
        logging.info(msg)

        try:
            run_abuild(
                context,
                apkbuild,
                pkg["pkgver"],
                channel,
                pkg_arch,
                cross,
                strict,
                force,
                src,
                bootstrap_stage,
            )
        except CommandFailedError as exception:
            raise BuildFailedError(f"Couldn't build {output}!") from exception
        finish(pkg["apkbuild"], channel, pkg_arch, output, buildchroot, strict)

    # Clear package cache for the next run
    _package_cache = {}

    if built_packages:
        logging.info("@YELLOW@=>@END@ @GREEN@Finished building packages")

    return list(built_packages)



# Copyright 2023 Mark Hargreaves, Luca Weiss
# SPDX-License-Identifier: GPL-3.0-or-later
import socket
import time
from pathlib import Path

import pmb.chroot
import pmb.helpers.run
from pmb.core import Chroot
from pmb.core.context import get_context
from pmb.helpers import logging
from pmb.helpers.exceptions import NonBugError


def start_nbd_server(device: str, replace: bool, ip: str = "172.16.42.2", port: int = 9999) -> None:
    """
    Start nbd server in chroot_native with pmOS rootfs.

    :param ip: IP address to serve nbd server for
    :param port: port of nbd server
    """
    chroot = Chroot.native()
    pmb.chroot.init(chroot)

    rootfs_path = Path("/mnt/pmbootstrap/netboot") / f"{device}.img"
    if not (chroot / rootfs_path).exists() or replace:
        rootfs_path2 = Path("/home/pmos/rootfs") / f"{device}.img"
        if not (chroot / rootfs_path2).exists():
            raise NonBugError(
                "The rootfs has not been generated yet, please run 'pmbootstrap install' first."
            )
        if replace and not pmb.helpers.cli.confirm(
            f"Are you sure you want to replace the rootfs for {device}?"
        ):
            return
        pmb.chroot.root(["cp", rootfs_path2, rootfs_path])
        logging.info(
            f"NOTE: Copied device image to {get_context().config.work}"
            f'/images_netboot/. The image will persist "pmbootstrap '
            f'zap" for your convenience. Use "pmbootstrap netboot '
            f'serve --help" for more options.'
        )

    pmb.chroot.apk.install(["nbd"], chroot)

    logging.info(f"Running nbd server for {device} on {ip} port {port}.")

    while True:
        logging.info("Waiting for postmarketOS device to appear...")

        # Try to bind to the IP ourselves before handing it to nbd-servere
        # This is purely to improve the UX as nbd-server just quits when it
        # cannot bind to an IP address.
        test_socket = socket.socket()
        while True:
            try:
                test_socket.bind((ip, 9998))
            except OSError as e:
                if e.errno != 99:  # Cannot assign requested address
                    raise
                # Wait a bit before retrying
                time.sleep(0.5)
                continue
            test_socket.close()
            break

        logging.info("Found postmarketOS device, serving image...")
        pmb.chroot.root(
            ["nbd-server", f"{ip}@{port}", rootfs_path, "-d"], check=False, disable_timeout=True
        )
        logging.info("nbd-server quit. Connection lost?")
        # On a reboot nbd-server will quit, but the IP address sticks around
        # for a bit longer, so wait.
        time.sleep(5)


def netboot(action_netboot: str, do_replace: bool) -> None:
    if action_netboot == "serve":
        device = get_context().config.device
        start_nbd_server(device, do_replace)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import pmb.helpers.run
import pmb.parse.apkindex
from pmb.aportgen.device import ask_for_architecture
from pmb.core.context import get_context
from pmb.helpers.exceptions import NonBugError
from pmb.parse.deviceinfo import Deviceinfo


def generate_apkbuild(
    pkgname: str,
    deviceinfo: Deviceinfo | None,
    patches: list[str],
    device_category: pmb.helpers.devices.DeviceCategory,
) -> None:
    device = "-".join(pkgname.split("-")[1:])
    arch = deviceinfo.arch if deviceinfo else ask_for_architecture()
    carch = arch.kernel_arch()

    makedepends = [
        "bash",
        "bc",
        "bison",
        "devicepkg-dev",
        "findutils",
        "flex",
        "openssl-dev",
        "perl",
    ]

    # Downstream kernel
    if device_category == pmb.helpers.devices.DeviceCategory.DOWNSTREAM:
        reference_url = "https://postmarketos.org/vendorkernel"

        outdir = '_outdir="out"\n'

        prepare = """
            default_prepare
            . downstreamkernel_prepare"""

        build = """
            unset LDFLAGS
            make O="$_outdir" ARCH="$_carch" CC="${CC:-gcc}" \\
                KBUILD_BUILD_VERSION="$((pkgrel + 1 ))-postmarketOS\""""

        package = """
            downstreamkernel_package "$builddir" "$pkgdir" "$_carch\" \\
                "$_flavor" "$_outdir\""""

        if deviceinfo:
            has_dtb = deviceinfo.header_version and deviceinfo.header_version >= 2
        else:
            has_dtb = pmb.helpers.cli.confirm(
                "Does the device use DTBs?", default=True, no_assumptions=True
            )

        if has_dtb:
            package += """
            make dtbs_install O="$_outdir" ARCH="$_carch" \\
                INSTALL_DTBS_PATH="$pkgdir\"/boot/dtbs"""

        if deviceinfo:
            has_qcdt = deviceinfo.bootimg_qcdt == "true"
        else:
            has_qcdt = pmb.helpers.cli.confirm(
                "Does the device use QCDT (see <https://wiki.postmarketos.org/wiki/QCDT>)?",
                default=False,
                no_assumptions=True,
            )

        if has_qcdt:
            build += """\n
            # Master DTB (deviceinfo_bootimg_qcdt)"""
            vendors = ["spreadtrum", "exynos", "other"]
            soc_vendor = pmb.helpers.cli.ask("SoC vendor", vendors, vendors[-1], complete=vendors)
            if soc_vendor == "spreadtrum":
                makedepends.append("dtbtool-sprd")
                build += """
            dtbTool-sprd -p "$_outdir/scripts/dtc/" \\
                -o "$_outdir/arch/$_carch/boot"/dt.img \\
                "$_outdir/arch/$_carch/boot/dts/\""""
            elif soc_vendor == "exynos":
                codename = "-".join(pkgname.split("-")[2:])
                makedepends.append("dtbtool-exynos")
                build += """
            dtbTool-exynos -o "$_outdir/arch/$_carch/boot"/dt.img \\
                $(find "$_outdir/arch/$_carch/boot/dts/\""""
                build += f" -name *{codename}*.dtb)"
            else:
                makedepends.append("dtbtool")
                build += """
            dtbTool -o "$_outdir/arch/$_carch/boot"/dt.img \\
                "$_outdir/arch/$_carch/boot/\""""
            package += """
            install -Dm644 "$_outdir/arch/$_carch/boot"/dt.img \\
                "$pkgdir"/boot/dt.img"""

    # Mainline kernel
    else:
        reference_url = None

        # Add LLVM dependencies
        makedepends += ["clang", "lld", "llvm"]

        outdir = ""

        prepare = """
            default_prepare
            cp -v "$srcdir/$_config" .config"""

        build = """
            unset LDFLAGS
            make ARCH="$_carch" LLVM=1 \\
                KBUILD_BUILD_VERSION="$((pkgrel + 1 ))-postmarketOS\""""

        package = """
            mkdir -p "$pkgdir"/boot
            make zinstall modules_install dtbs_install \\
                ARCH="$_carch" \\
                LLVM=1 \\
                INSTALL_MOD_STRIP=1 \\
                INSTALL_PATH="$pkgdir"/boot \\
                INSTALL_MOD_PATH="$pkgdir" \\
                INSTALL_DTBS_PATH="$pkgdir/boot/dtbs"

            install -D "$builddir"/include/config/kernel.release \\
                "$pkgdir/usr/share/kernel/$_flavor/kernel.release\""""

    makedepends.sort()
    makedepends_fmt = ("\n" + " " * 12).join(makedepends)
    patches_str = ("\n" + " " * 12).join(patches)
    reference_str = " " * 8 + f"# Reference: <{reference_url}>\n" if reference_url else ""
    content = f"""{reference_str}\
        # Kernel config based on: arch/{carch}/configs/(CHANGEME!)

        maintainer=""
        pkgname={pkgname}
        pkgver=3.x.x
        pkgrel=0
        pkgdesc="{deviceinfo.name if deviceinfo else "(CHANGEME!)"} kernel fork"
        arch="{arch}"
        _carch="{carch}"
        _flavor="{device}"
        url="https://kernel.org"
        license="GPL-2.0-only"
        options="!strip !check !tracedeps pmb:cross-native"
        makedepends="
            {makedepends_fmt}
        "

        # Source
        _repository="(CHANGEME!)"
        _commit="ffffffffffffffffffffffffffffffffffffffff"
        _config="config-$_flavor.$arch"
        source="
            $pkgname-$_commit.tar.gz::https://github.com/(CHANGEME!)/$_repository/archive/$_commit.tar.gz
            $_config
            {patches_str}
        "
        builddir="$srcdir/$_repository-$_commit"
        {outdir}
        prepare() {{{prepare}
        }}

        build() {{{build}
        }}

        package() {{{package}
        }}

        sha512sums="(run 'pmbootstrap checksum {pkgname}' to fill)"
        """

    # Write the file
    with (get_context().config.work / "aportgen/APKBUILD").open("w", encoding="utf-8") as hndl:
        for line in content.rstrip().split("\n"):
            hndl.write(line[8:].replace(" " * 4, "\t") + "\n")


def generate(pkgname: str, device_category: pmb.helpers.devices.DeviceCategory) -> None:
    device = "-".join(pkgname.split("-")[1:])
    try:
        deviceinfo = pmb.parse.deviceinfo(device)
    except NonBugError:  # device not found
        deviceinfo = None
    work = get_context().config.work

    pmb.helpers.run.user(["mkdir", "-p", work / "aportgen"])

    # Symlink commonly used patches
    if device_category == pmb.helpers.devices.DeviceCategory.DOWNSTREAM:
        patches = [
            "gcc7-give-up-on-ilog2-const-optimizations.patch",
            "gcc8-fix-put-user.patch",
            "gcc10-extern_YYLOC_global_declaration.patch",
            "kernel-use-the-gnu89-standard-explicitly.patch",
        ]
        for patch in patches:
            pmb.helpers.run.user(
                ["ln", "-s", "../../.shared-patches/linux/" + patch, (work / "aportgen" / patch)]
            )
    else:
        patches = []

    generate_apkbuild(pkgname, deviceinfo, patches, device_category)



# Copyright 2024 Caleb Connolly
# SPDX-License-Identifier: GPL-3.0-or-later

import glob
import os
from collections.abc import Generator
from pathlib import Path

import pmb.config
from pmb.core.context import get_context
from pmb.meta import Cache
from pmb.types import WithExtraRepos


@Cache("with_extra_repos")
def pkgrepo_paths(with_extra_repos: WithExtraRepos = "default") -> list[Path]:
    config = get_context().config
    paths = [Path(x) for x in config.aports]
    if not paths:
        raise RuntimeError("No package repositories specified?")

    with_systemd = False

    match with_extra_repos:
        case "disabled":
            return paths
        case "enabled":
            with_systemd = True
        case "default":
            with_systemd = pmb.config.is_systemd_selected(config)

    out_paths = []
    for p in paths:
        # This isn't very generic, but we don't plan to add new extra-repos...
        if (p / "extra-repos/systemd").is_dir() and with_systemd:
            out_paths.append(p / "extra-repos/systemd")
        out_paths.append(p)

    return out_paths


@Cache()
def pkgrepo_default_path() -> Path:
    return pkgrepo_paths(with_extra_repos="disabled")[0]


def pkgrepo_names(with_extra_repos: WithExtraRepos = "default") -> list[str]:
    """
    Return a list of all the package repository names. We REQUIRE
    that the last repository is "pmaports", though the directory
    may be named differently. So we hardcode the name here.
    """
    names = [aports.name for aports in pkgrepo_paths(with_extra_repos)]
    names[-1] = "pmaports"
    return names


def pkgrepo_name(path: Path) -> str:
    """
    Return the name of the package repository with the given path. This
    MUST be used instead of "path.name" as we need special handling
    for the pmaports repository.
    """
    if path == get_context().config.aports[-1]:
        return "pmaports"

    return path.name


def pkgrepo_path(name: str) -> Path:
    """Return the absolute path to the package repository with the given name."""
    # The pmaports repo is always last, and we hardcode the name.
    if name == "pmaports":
        return get_context().config.aports[-1]

    for aports in pkgrepo_paths():
        if aports.name == name:
            return aports
    raise RuntimeError(f"aports '{name}' not found")


def pkgrepo_name_from_subdir(subdir: Path) -> str:
    """
    Return the name of the package repository for the given directory.
    e.g. "pmaports" for "$WORKDIR/pmaports/main/foobar"
    """
    for aports in pkgrepo_paths():
        if subdir.is_relative_to(aports):
            return pkgrepo_name(aports)
    raise RuntimeError(f"aports subdir '{subdir}' not found")


def pkgrepo_glob_one(path: str) -> Path | None:
    """
    Search for the file denoted by path in all aports repositories.
    path can be a glob.
    """
    for aports in pkgrepo_paths():
        g = glob.glob(os.path.join(aports, path), recursive=True)
        if not g:
            continue

        if len(g) != 1:
            raise RuntimeError(f"{path} found multiple matches in {aports}")
        if g:
            return aports / g[0]

    return None


def pkgrepo_iglob(path: str, recursive: bool = False) -> Generator[Path, None, None]:
    """Yield each matching glob over each aports repository."""
    for repo in pkgrepo_paths():
        for g in glob.iglob(os.path.join(repo, path), recursive=recursive):
            pdir = Path(g)
            # Skip extra-repos when not parsing the extra-repo itself
            if "extra-repos" not in repo.parts and "extra-repos" in pdir.parts:
                continue
            yield pdir


def pkgrepo_iter_package_dirs(
    with_extra_repos: WithExtraRepos = "default",
) -> Generator[Path, None, None]:
    """
    Yield each matching glob over each aports repository.
    Detect duplicates within the same aports repository but otherwise
    ignore all but the first. This allows for overriding packages.
    """
    seen: dict[str, list[str]] = {a: [] for a in pkgrepo_names(with_extra_repos)}

    for repo in pkgrepo_paths(with_extra_repos):
        for root, dirs, files in os.walk(repo, followlinks=True):
            # Skip hidden directories and common non-package directories
            dirs[:] = [d for d in dirs if not d.startswith(".")]

            if "APKBUILD" in files:
                pdir = Path(root)

                # Skip extra-repos when not parsing the extra-repo itself
                if "extra-repos" not in repo.parts and "extra-repos" in pdir.parts:
                    continue

                pkg = pdir.name
                if pkg in seen[pkgrepo_name(repo)]:
                    raise RuntimeError(
                        f"Package {pkg} found in multiple aports "
                        "subfolders. Please put it only in one folder."
                    )
                if pkg in [x for li in seen.values() for x in li]:
                    continue

                seen[pkgrepo_name(repo)].append(pkg)
                yield pdir

                # Don't traverse subdirectories of a package directory
                dirs.clear()


def pkgrepo_relative_path(path: Path) -> tuple[Path, Path]:
    """Return the path relative to the first aports repository."""
    for aports in pkgrepo_paths():
        # Cheeky hack to let us jump from an extra-repo to the
        # main aports repository it's a part of
        if path != aports and path.is_relative_to(aports):
            return aports, path.relative_to(aports)
    raise RuntimeError(f"Path '{path}' not in aports repositories")



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
from pmb.parse._apkbuild import (
    apkbuild as apkbuild,
    function_body as function_body,
)
from pmb.parse.arguments import (
    arguments as arguments,
    arguments_flasher as arguments_flasher,
    arguments_install as arguments_install,
    get_parser as get_parser,
)
from pmb.parse.binfmt_info import binfmt_info as binfmt_info
from pmb.parse.bootimg import bootimg as bootimg
from pmb.parse.cpuinfo import arm_big_little_first_group_ncpus as arm_big_little_first_group_ncpus
from pmb.parse.deviceinfo import deviceinfo as deviceinfo
from pmb.parse.kconfig import check as check



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import contextlib
import datetime
import os
import re
import readline
import sys
from collections.abc import KeysView
from typing import Any

import pmb.config
from pmb.core.context import get_context
from pmb.helpers import logging


class ReadlineTabCompleter:
    """Store intermediate state for completer function."""

    def __init__(self, options: KeysView[str] | dict[str, Any] | list[str]) -> None:
        """:param options: list of possible completions."""
        self.options = sorted(options)
        self.matches: list[str] = []

    def completer_func(self, input_text: str, iteration: int) -> str | None:
        """
        :param input_text: text that shall be autocompleted
        :param iteration: how many times "tab" was hit
        """
        # First time: build match list
        if iteration == 0:
            if input_text:
                self.matches = [s for s in self.options if s and s.startswith(input_text)]
            else:
                self.matches = self.options[:]

        # Return the N'th item from the match list, if we have that many.
        if iteration < len(self.matches):
            return self.matches[iteration]
        return None


def ask(
    question: str = "Continue?",
    choices: list[str] | None = ["y", "n"],
    default: int | str | None = "n",
    lowercase_answer: bool | None = True,
    validation_regex: str | None = None,
    complete: KeysView[str] | dict[str, Any] | list[str] | None = None,
) -> str:
    """
    Ask a question on the terminal.

    :param question: display prompt
    :param choices: short list of possible answers, displayed after prompt if set
    :param default: default value to return if user doesn't input anything
    :param lowercase_answer: if True, convert return value to lower case
    :param validation_regex: if set, keep asking until regex matches
    :param complete: set to a list to enable tab completion
    """
    styles = pmb.config.styles

    while True:
        date = datetime.datetime.now().strftime("%H:%M:%S")
        line = question
        if choices:
            line += f" ({str.join('/', choices)})"
        if default:
            line += f" [{default}]"
        line_color = f"[{date}] {styles['BOLD']}{line}{styles['END']}"
        line = f"[{date}] {line}"

        if complete:
            readline.parse_and_bind("tab: complete")
            delims = readline.get_completer_delims()
            if "-" in delims:
                delims = delims.replace("-", "")
                readline.set_completer_delims(delims)
            readline.set_completer(ReadlineTabCompleter(complete).completer_func)

        ret = input(f"{line_color}: ")

        # Stop completing (question is answered)
        if complete:
            # set_completer(None) would use the default file system completer
            readline.set_completer(lambda text, state: None)

        if lowercase_answer:
            ret = ret.lower()
        if ret == "":
            ret = str(default)

        pmb.helpers.logging.logfd.write(f"{line}: {ret}\n")
        pmb.helpers.logging.logfd.flush()

        # Validate with regex
        if not validation_regex:
            return ret

        if re.fullmatch(validation_regex, ret):
            return ret

        logging.fatal(
            "ERROR: Input did not pass validation (regex: "
            + validation_regex
            + "). Please try again."
        )


def confirm(
    question: str = "Continue?", default: bool = False, no_assumptions: bool = False
) -> bool:
    """
    Convenience wrapper around ask for simple yes-no questions with validation.

    :param no_assumptions: ask for confirmation, even if "pmbootstrap -y' is set
    :returns: True for "y", False for "n"
    """
    default_str = "y" if default else "n"
    if get_context().assume_yes and not no_assumptions:
        logging.info(question + " (y/n) [" + default_str + "]: y")
        return True
    answer = ask(question, ["y", "n"], default_str, True, "(y|n)")
    return answer == "y"


def progress_print(progress: float) -> None:
    """
    Print a snapshot of a progress bar to STDOUT.

    Call progress_flush to end  printing progress and clear the line. No output is printed in
    non-interactive mode.

    :param progress: completion percentage as a number between 0 and 1
    """
    width = 79
    with contextlib.suppress(OSError):
        width = os.get_terminal_size().columns - 6
    chars = int(width * progress)
    filled = "\u2588" * chars
    empty = " " * (width - chars)
    percent = int(progress * 100)
    if pmb.config.is_interactive and not get_context().details_to_stdout:
        sys.stdout.write(f"\u001b7{percent:>3}% {filled}{empty}")
        sys.stdout.flush()
        sys.stdout.write("\u001b8\u001b[0K")


def progress_flush() -> None:
    """
    Finish printing a progress bar.

    This will erase the line. Does nothing in non-interactive mode.
    """
    if pmb.config.is_interactive and not get_context().details_to_stdout:
        sys.stdout.flush()



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import configparser
import os
from pathlib import Path, PosixPath

from pmb.core import Config
from pmb.helpers import logging


def load(path: Path) -> Config:
    config = Config()

    cfg = configparser.ConfigParser()
    if os.path.isfile(path):
        cfg.read(path)

    if "pmbootstrap" not in cfg:
        cfg["pmbootstrap"] = {}
    if "providers" not in cfg:
        cfg["providers"] = {}

    for key in Config.__dict__:
        if key == "providers":
            setattr(config, key, cfg["providers"])
        if key == "mirrors" and key in cfg:
            for subkey in Config.mirrors:
                if subkey in cfg["mirrors"]:
                    setattr(config, f"mirrors.{subkey}", cfg["mirrors"][subkey])
        # default values won't be set in the config file
        if key not in cfg["pmbootstrap"]:
            continue
        # Convert strings to paths
        elif type(getattr(Config, key)) is PosixPath:
            setattr(config, key, Path(cfg["pmbootstrap"][key]))
        # Yeah this really sucks and there isn't a better way to do it without external
        # libraries
        elif isinstance(getattr(Config, key), list) and isinstance(
            getattr(Config, key)[0], PosixPath
        ):
            value = cfg["pmbootstrap"][key]
            if not value:
                setattr(config, key, value)
            else:
                setattr(config, key, [Path(p) for p in value.split(",")])
        elif isinstance(getattr(Config, key), bool):
            setattr(config, key, cfg["pmbootstrap"][key].lower() == "true")
        elif key in cfg["pmbootstrap"]:
            setattr(config, key, cfg["pmbootstrap"][key])

    return config


def serialize(config: Config, skip_defaults: bool = True) -> configparser.ConfigParser:
    """
    Serialize the config object into a ConfigParser to write it out
    in the pmbootstrap_v3.cfg INI format.

    :param config: The config object to serialize
    :param skip_defaults: Skip writing out default values
    """
    cfg = configparser.ConfigParser()
    cfg["pmbootstrap"] = {}
    cfg["providers"] = {}
    cfg["mirrors"] = {}

    # .keys() flat maps dictionaries like config.mirrors with
    # dotted notation
    for key in Config.keys():  # noqa: SIM118
        # If the default value hasn't changed then don't write out,
        # this makes it possible to update the default, otherwise
        # we wouldn't be able to tell if the user overwrote it.
        if skip_defaults and Config.get_default(key) == getattr(config, key):
            continue
        if key == "providers":
            cfg["providers"] = config.providers
        elif key.startswith("mirrors."):
            key_ = key.split(".")[1]
            cfg["mirrors"][key_] = getattr(config, key)
        # Convert strings to paths
        elif type(getattr(Config, key)) is PosixPath:
            cfg["pmbootstrap"][key] = str(getattr(config, key))
        elif isinstance(getattr(Config, key), list) and isinstance(
            getattr(Config, key)[0], PosixPath
        ):
            cfg["pmbootstrap"][key] = ",".join(os.fspath(p) for p in getattr(config, key))
        elif isinstance(getattr(Config, key), bool):
            cfg["pmbootstrap"][key] = str(getattr(config, key))
        else:
            cfg["pmbootstrap"][key] = str(getattr(config, key))

    return cfg


# FIXME: we should have distinct Config and ConfigFile types
def save(output: Path, config: Config) -> None:
    """
    Save the config object to the specified path.

    IMPORTANT: The global config (available via get_context().config)
    has invocation arguments merged into it. Do NOT call save() with
    the global config object.
    """
    logging.debug(f"Save config: {output}")
    output.parent.mkdir(parents=True, exist_ok=True)
    output.touch(0o600, exist_ok=True)

    cfg = serialize(config)

    with output.open("w") as handle:
        cfg.write(handle)



# Copyright 2026 Stefan Hansson, Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
from collections.abc import Collection
from pathlib import Path

import pmb.build.checksum
from pmb.core import Chroot
from pmb.helpers import logging


def checksum(packages: Collection[str], do_changed: bool, do_verify: bool) -> None:
    def get_relevant_packages() -> Collection[str]:
        if do_changed:
            return pmb.helpers.git.get_changed_packages()
        elif packages:
            return packages
        else:
            return {Path.cwd().name}

    pmb.chroot.init(Chroot.native())

    packages = get_relevant_packages()

    for package in packages:
        if do_verify:
            pmb.build.checksum.verify(package)
        else:
            pmb.build.checksum.update(package)

    if not packages:
        # We should only ever reach this if the --changed argument is used as
        # otherwise at least one package must be specified in the arguments.
        logging.info("NOTE: No changed packages detected, not updating any checksums")



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import json
import time
from pathlib import Path

import pmb.chroot
import pmb.helpers.mount
import pmb.helpers.run
from pmb.core import Chroot
from pmb.core.context import get_context
from pmb.helpers import logging
from pmb.types import PathString


def init() -> None:
    if not Path("/sys/module/loop").is_dir():
        pmb.helpers.run.root(["modprobe", "loop"])
    for loopdevice in Path("/dev/").glob("loop*"):
        if loopdevice.is_dir():
            continue
        pmb.helpers.mount.bind_file(loopdevice, Chroot.native() / loopdevice)


def mount(img_path: Path, _sector_size: int | None = None) -> Path:
    """:param img_path: Path to the img file inside native chroot."""
    logging.debug(f"(native) mount {img_path} (loop)")

    # Try to mount multiple times (let the kernel module initialize #1594)
    for i in range(5):
        # Retry
        if i > 0:
            logging.debug("loop module might not be initialized yet, retry in one second...")
            time.sleep(1)

        # Mount and return on success
        init()

        sector_size = None
        if _sector_size:
            sector_size = str(_sector_size)
        sector_size = sector_size or pmb.parse.deviceinfo().rootfs_image_sector_size

        losetup_cmd: list[PathString] = ["losetup", "-f", "-P", img_path]
        if sector_size:
            losetup_cmd += ["-b", str(int(sector_size))]

        pmb.chroot.root(losetup_cmd, check=False)

        try:
            return device_by_back_file(img_path)
        except RuntimeError:
            if i == 4:
                raise

    raise AssertionError("This should never be reached")


def device_by_back_file(back_file: Path) -> Path:
    """Get the /dev/loopX device that points to a specific image file."""
    # Get list from losetup
    losetup_output = pmb.chroot.root(["losetup", "--json", "--list"], output_return=True)
    if not losetup_output:
        raise RuntimeError("losetup failed")

    # Find the back_file
    losetup = json.loads(losetup_output)
    for loopdevice in losetup["loopdevices"]:
        if loopdevice["back-file"] is not None and Path(loopdevice["back-file"]) == back_file:
            return Path(loopdevice["name"])
    raise RuntimeError(f"Failed to find loop device for {back_file}")


def umount(img_path: Path) -> None:
    """:param img_path: Path to the img file inside native chroot."""
    device: Path
    try:
        device = device_by_back_file(img_path)
    except RuntimeError:
        return
    logging.debug(f"(native) umount {device}")
    pmb.chroot.root(["losetup", "-d", device])


def detach_all() -> None:
    """Detach all loop devices used by pmbootstrap"""
    losetup_output = pmb.helpers.run.root(["losetup", "--json", "--list"], output_return=True)
    if not losetup_output:
        return
    losetup = json.loads(losetup_output)
    work = get_context().config.work
    for loopdevice in losetup["loopdevices"]:
        if Path(loopdevice["back-file"]).is_relative_to(work):
            pmb.helpers.run.root(["kpartx", "-d", loopdevice["name"]], check=False)
            pmb.helpers.run.root(["losetup", "-d", loopdevice["name"]])
    return



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import json

import pmb.helpers.repo_missing
from pmb.core.arch import Arch
from pmb.helpers import logging


def repo_missing(arch: Arch | None, built: bool) -> None:
    if arch is None:
        raise AssertionError
    if built:
        logging.warning(
            "WARNING: --built is deprecated (bpo#148: this warning is expected on build.postmarketos.org for now)"
        )
    missing = pmb.helpers.repo_missing.generate(arch)
    print(json.dumps(missing, indent=4))



# Copyright 2024 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import configparser
import os
import re
from enum import Enum
from glob import glob
from pathlib import Path
from typing import Final
from urllib.parse import urlparse

import pmb.config
import pmb.helpers.pmaports
import pmb.helpers.run
from pmb.core.context import get_context
from pmb.core.pkgrepo import pkgrepo_default_path, pkgrepo_name, pkgrepo_path
from pmb.helpers import logging
from pmb.helpers.exceptions import NonBugError
from pmb.meta import Cache
from pmb.types import PathString, RunOutputTypeDefault

re_branch_aports = re.compile(r"^\d+\.\d\d+-stable$")
re_branch_pmaports = re.compile(r"^v\d\d\.\d\d$")


def get_path(name_repo: str) -> Path:
    """
    Get the path to the repository.

    The path is either the default one in the work dir, or a user-specified one in args.

    :returns: full path to repository
    """
    if name_repo == "aports_upstream":
        return get_context().config.work / "cache_git" / name_repo
    return pkgrepo_path(name_repo)


def clone(name_repo: str, do_shallow_clone: bool = False) -> None:
    """
    Clone a git repository to $WORK/cache_git/$name_repo.

    (or to the overridden path set in args, as with ``pmbootstrap --aports``).

    :param name_repo: short alias used for the repository name, from pmb.config.git_repos
        (e.g. "aports_upstream", "pmaports")
    :param do_shallow_clone: Whether to do a shallow (--depth=1) clone of the given repository.
    """
    # Check for repo name in the config
    if name_repo not in pmb.config.git_repos:
        raise ValueError("No git repository configured for " + name_repo)

    path = get_path(name_repo)
    if not path.exists():
        # Build git command
        url = pmb.config.git_repos[name_repo][0]
        command = ["git", "clone"]
        if do_shallow_clone:
            command.append("--depth=1")
        command += [url, str(path)]

        # Create parent dir and clone
        logging.info(f"Clone git repository: {url}")
        (get_context().config.work / "cache_git").mkdir(exist_ok=True)
        pmb.helpers.run.user(command, output=RunOutputTypeDefault.STDOUT)

    # FETCH_HEAD does not exist after initial clone. Create it, so
    # is_outdated() can use it.
    fetch_head = path / ".git/FETCH_HEAD"
    if not fetch_head.exists():
        open(fetch_head, "w").close()


def rev_parse(
    path: Path, revision: str = "HEAD", extra_args: list = [], silent: bool = False
) -> str:
    """
    Run "git rev-parse" in a specific repository dir.

    :param path: to the git repository
    :param extra_args: additional arguments for ``git rev-parse``. Pass
        ``--abbrev-ref`` to get the branch instead of the commit, if possible.
    :returns: commit string like "90cd0ad84d390897efdcf881c0315747a4f3a966"
        or (with ``--abbrev-ref``): the branch name, e.g. "main"
    """
    command = ["git", "rev-parse", *extra_args, revision]
    rev = pmb.helpers.run.user_output(
        command, path, output=RunOutputTypeDefault.NULL if silent else RunOutputTypeDefault.LOG
    )
    return rev.rstrip()


def can_fast_forward(path: Path, branch_upstream: str, branch: str = "HEAD") -> bool:
    command = ["git", "merge-base", "--is-ancestor", branch, branch_upstream]
    ret = pmb.helpers.run.user(command, path, check=False)
    if ret == 0:
        return True
    elif ret == 1:
        return False
    else:
        raise RuntimeError("Unexpected exit code from git: " + str(ret))


def clean_worktree(path: Path, silent: bool = False) -> bool:
    """Check if there are not any modified files in the git dir."""
    command = ["git", "status", "--porcelain"]
    return (
        pmb.helpers.run.user_output(
            command, path, output=RunOutputTypeDefault.NULL if silent else RunOutputTypeDefault.LOG
        )
        == ""
    )


def list_remotes(aports: Path) -> list[str]:
    command = ["git", "remote", "-v"]
    output = pmb.helpers.run.user_output(command, aports, output=RunOutputTypeDefault.NULL)
    return output.splitlines()


def remote_to_name_and_clean_url(remote_line: str) -> tuple[str, str]:
    remote_name, url_part = remote_line.split("\t", 1)

    if url_part.startswith("https://"):
        # GitLab CI adds an authentication segment to the URL, which needs to be removed to be able
        # to match against the expected URLs later.
        parsed = urlparse(url_part)
        clean_url = f"{parsed.scheme}://{parsed.hostname}{parsed.path}"
    else:
        clean_url = url_part

    return (remote_name, clean_url)


def get_upstream_remote(aports: Path) -> str:
    """
    Find the remote, which matches the git URL from the config.

    Usually "origin", but the user may have set up their git repository differently.
    """
    name_repo = pkgrepo_name(aports)
    if name_repo not in pmb.config.git_repos:
        logging.warning(f"WARNING: can't determine remote for {name_repo}, using 'origin'")
        return "origin"
    urls = pmb.config.git_repos[name_repo]
    lines = list_remotes(aports)
    for line in lines:
        remote_name, clean_url = remote_to_name_and_clean_url(line)

        if any(u.lower() in clean_url.lower() for u in urls):
            return remote_name

    # Fallback to old URLs, in case the migration was not done yet
    if name_repo == "pmaports":
        urls_outdated = OUTDATED_GIT_REMOTES_HTTP + OUTDATED_GIT_REMOTES_SSH
        for line in lines:
            if any(u in line.lower() for u in urls_outdated):
                logging.warning("WARNING: pmaports has an outdated remote URL")
                return line.split("\t", 1)[0]

    raise RuntimeError(
        f"{name_repo}: could not find remote name for any URL '{urls}' in git repository: {aports}"
    )


class RemoteType(Enum):
    FETCH = "fetch"
    PUSH = "push"

    @staticmethod
    def from_git_output(git_remote_type: str) -> "RemoteType":
        match git_remote_type:
            case "(fetch)":
                return RemoteType.FETCH
            case "(push)":
                return RemoteType.PUSH
            case _:
                raise ValueError(f'Unknown remote type "{git_remote_type}"')


def set_remote_url(repo: Path, remote_name: str, remote_url: str, remote_type: RemoteType) -> None:
    command: list[PathString] = [
        "git",
        "-C",
        repo,
        "remote",
        "set-url",
        remote_name,
        remote_url,
        "--push" if remote_type == RemoteType.PUSH else "--no-push",
    ]

    pmb.helpers.run.user(command, output=RunOutputTypeDefault.STDOUT)


# Intentionally lower case for case-insensitive comparison
OUTDATED_GIT_REMOTES_HTTP: Final[list[str]] = ["https://gitlab.com/postmarketos/pmaports.git"]
OUTDATED_GIT_REMOTES_SSH: Final[list[str]] = ["git@gitlab.com:postmarketos/pmaports.git"]


def migrate_upstream_remote() -> None:
    """Migrate pmaports git remote URL from gitlab.com to gitlab.postmarketos.org."""
    repo = pkgrepo_default_path()
    repo_name = repo.parts[-1]
    lines = list_remotes(repo)

    current_git_remote_http: Final[str] = pmb.config.git_repos[repo_name][0]
    current_git_remote_ssh: Final[str] = pmb.config.git_repos[repo_name][1]

    for line in lines:
        if not line:
            continue  # Skip empty line at the end.

        remote_name, remote_url, remote_type_raw = line.split()
        remote_type = RemoteType.from_git_output(remote_type_raw)

        if remote_url.lower() in OUTDATED_GIT_REMOTES_HTTP:
            new_remote = current_git_remote_http
        elif remote_url.lower() in OUTDATED_GIT_REMOTES_SSH:
            new_remote = current_git_remote_ssh
        else:
            new_remote = None

        if new_remote:
            logging.info(
                f"Migrating to new {remote_type.value} URL (from {remote_url} to {new_remote})"
            )
            set_remote_url(repo, remote_name, current_git_remote_http, remote_type)


@Cache("aports")
def parse_channels_cfg(aports: Path) -> dict:
    """
    Parse channels.cfg from pmaports.git, origin/main branch.

    Reference: https://postmarketos.org/channels.cfg

    :returns: dict like: {"meta": {"recommended": "edge"},
        "channels": {"edge": {"description": ...,
        "branch_pmaports": ...,
        "branch_aports": ...,
        "mirrordir_alpine": ...},
        ...}}
    """
    # Read with configparser
    cfg = configparser.ConfigParser()
    override = os.environ.get("PMB_CHANNELS_CFG")

    if override:
        if not os.path.exists(override):
            raise NonBugError(f"Override from PMB_CHANNELS_CFG not found: {override}")
        try:
            cfg.read(override)
        except configparser.MissingSectionHeaderError as exception:
            raise RuntimeError(
                f"Failed to read channels.cfg from '{override}' (as set in environment by PMB_CHANNELS_CFG)"
            ) from exception
    else:
        remote = get_upstream_remote(aports)
        command = ["git", "show", f"{remote}/main:channels.cfg"]
        stdout = pmb.helpers.run.user_output(
            command, aports, output=RunOutputTypeDefault.NULL, check=False
        )
        try:
            cfg.read_string(stdout)
        except configparser.MissingSectionHeaderError as exception:
            logging.info(
                "NOTE: fix this by fetching your pmaports.git, e.g. with 'pmbootstrap pull'"
            )
            raise RuntimeError(
                f"Failed to read channels.cfg from '{remote}/main' branch of your local pmaports clone"
            ) from exception

    # Meta section
    ret: dict[str, dict[str, str | dict[str, str]]] = {"channels": {}}
    ret["meta"] = {"recommended": cfg.get("channels.cfg", "recommended")}

    # Channels
    for channel in cfg.sections():
        if channel == "channels.cfg":
            continue  # meta section

        channel_new = pmb.helpers.pmaports.get_channel_new(channel)

        ret["channels"][channel_new] = {}
        for key in ["description", "branch_pmaports", "branch_aports", "mirrordir_alpine"]:
            value = cfg.get(channel, key)
            # FIXME: how to type this properly??
            ret["channels"][channel_new][key] = value  # type: ignore[index]

    return ret


def branch_looks_official(repo: Path, branch: str) -> bool:
    """
    Check if a given branch follows the patterns of official branches in
       pmaports or aports.

    :returns: True if it looks official, False otherwise
    """
    if branch == "main":
        return True
    if repo.parts[-1] == "pmaports":
        if re_branch_pmaports.match(branch):
            return True
    else:
        if re_branch_aports.match(branch):
            return True
    return False


def pull(repo_name: str) -> int:
    """
    Check if on official branch and essentially try ``git pull --ff-only``.

    Instead of really doing ``git pull --ff-only``, do it in multiple steps
    (``fetch, merge --ff-only``), so we can display useful messages depending
    on which part fails.

    :returns: integer, >= 0 on success, < 0 on error
    """
    repo = get_path(repo_name)

    # Skip if repo wasn't cloned
    if not os.path.exists(repo):
        logging.debug(repo_name + ": repo was not cloned, skipping pull!")
        return 1

    # Skip if not on official branch
    branch = rev_parse(repo, extra_args=["--abbrev-ref"])
    msg_start = f"{repo_name} (branch: {branch}):"
    if not branch_looks_official(repo, branch):
        if repo.parts[-1] == "pmaports":
            official_looking_branches = "main, v24.06, …"
        else:
            official_looking_branches = "master, 3.20-stable, …"
        logging.warning(
            f"{msg_start} not on one of the official branches"
            f" ({official_looking_branches}), skipping pull!"
        )
        return -1

    # Skip if workdir is not clean
    if not clean_worktree(repo):
        logging.warning(msg_start + " workdir is not clean, skipping pull!")
        return -2

    # Skip if branch is tracking different remote
    branch_upstream = get_upstream_remote(repo) + "/" + branch
    remote_ref = rev_parse(repo, branch + "@{u}", ["--abbrev-ref"])
    if remote_ref != branch_upstream:
        logging.warning(
            f"{msg_start} is tracking unexpected remote branch '{remote_ref}' instead"
            f" of '{branch_upstream}'"
        )
        return -3

    # Fetch (exception on failure, meaning connection to server broke)
    logging.info(msg_start + " git pull --ff-only")
    if not get_context().offline:
        pmb.helpers.run.user(["git", "fetch"], repo)

    # Skip if already up to date
    if rev_parse(repo, branch) == rev_parse(repo, branch_upstream):
        logging.info(msg_start + " already up to date")
        return 2

    # Skip if we can't fast-forward
    if not can_fast_forward(repo, branch_upstream):
        logging.warning(
            f"{msg_start} can't fast-forward to {branch_upstream}, looks like you changed"
            " the git history of your local branch. Skipping pull!"
        )
        return -4

    # Fast-forward now (should not fail due to checks above, so it's fine to
    # throw an exception on error)
    command = ["git", "merge", "--ff-only", branch_upstream]
    pmb.helpers.run.user(command, repo, RunOutputTypeDefault.STDOUT)
    return 0


def get_topdir(repo: Path) -> Path:
    """
    Get top-dir of git repo.

    :returns: the top dir of the git repository
    """
    res = pmb.helpers.run.user(
        ["git", "rev-parse", "--show-toplevel"], repo, output_return=True, check=False
    )
    if not isinstance(res, str):
        raise RuntimeError("Not a git repository: " + str(repo))
    return Path(res.strip())


def get_files(repo: Path, include_dot_git_dir: bool = False) -> list[str]:
    """
    Get all files inside a git repository, that are either already in the git tree or are not in gitignore.

    Do not list deleted files. To be used for creating a tarball of the git repository.

    :param path: top dir of the git repository
    :param include_dot_git_dir: Also include .git

    :returns: all files in a git repository as list, relative to path
    """
    ret = []
    files = pmb.helpers.run.user_output(["git", "ls-files"], repo).split("\n")
    files += pmb.helpers.run.user_output(
        ["git", "ls-files", "--exclude-standard", "--other"], repo
    ).split("\n")
    if include_dot_git_dir:
        files += glob("./.git/**/*", root_dir=repo, recursive=True)
    for file in files:
        if os.path.exists(f"{repo}/{file}"):
            ret += [file]

    return ret


def get_changed_files() -> set[Path]:
    """Get all changed files.

    Provides a set of all file paths that have uncommitted changes—both staged
    and unstaged—in the default package repository's git repository (usually
    pmaports).

    :returns: Set of changed file paths.
    """
    # Changed files
    ret = set()
    for file_string in pmb.helpers.run.user_output(
        # Diff against HEAD so we get both staged and unstaged changes.
        ["git", "diff", "--name-only", "HEAD"],
        working_dir=pkgrepo_default_path(),
    ).splitlines():
        file = Path(file_string)

        if (pkgrepo_default_path() / file).exists():
            ret.add(file)
    return ret


def _is_path_hidden(maybe_hidden: Path) -> bool:
    """Check whether a path would be considered hidden on Unix systems.

    Attempts to determine whether a path would be considered hidden on a Unix
    system.

    :param maybe_hidden: Path to analyse.
    :returns: True if the path is to be considered hidden, False otherwise.
    """
    return any(part[0] == "." for part in maybe_hidden.parts)


def get_changed_packages() -> set[str]:
    """Get all changed packages.

    Provides a set of the names of all packages that have uncommitted
    changes—both staged and unstaged—in the default package repository's git
    repository (usually pmaports).

    :returns: Set of the names of changed packages.
    """
    ret = set()
    for file in get_changed_files():
        # Skip files:
        # * in the root dir of pmaports (i.e. files with only one parent like ./README.md, which
        #   has the parent ".")
        # * path with a dot (e.g. .ci/, device/.shared-patches/)
        if len(file.parents) < 2 or _is_path_hidden(file):
            continue

        dirname = file.parent
        if file.name != "APKBUILD":
            # Walk up directories until we (eventually) find the package
            # the file belongs to (could be in a subdirectory of a package)
            while (
                dirname != dirname.parent
                and not (pkgrepo_default_path() / dirname / "APKBUILD").exists()
            ):
                dirname = dirname.parent

            # Unable to find APKBUILD the file belong to
            if dirname == dirname.parent:
                # ... maybe the package was deleted entirely?
                if not (pkgrepo_default_path() / file).exists():
                    continue

                # Weird, file does not belong to any package?
                continue
        elif not (pkgrepo_default_path() / file).exists():
            continue  # APKBUILD was deleted

        ret.add(dirname.name)

    return ret



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import pmb.chroot
import pmb.chroot.apk
import pmb.parse.apkindex
from pmb.core import Chroot
from pmb.core.context import get_context
from pmb.helpers import logging


def package_provider(
    pkgname: str, pkgnames_install: list[str], suffix: Chroot = Chroot.native()
) -> pmb.core.apkindex_block.ApkindexBlock | None:
    """
    :param pkgnames_install: packages to be installed
    :returns: ApkindexBlock object or None (no provider found)
    """
    # Get all providers
    arch = suffix.arch
    providers = pmb.parse.apkindex.providers(pkgname, arch, False)

    # 0. No provider
    if len(providers) == 0:
        return None

    # 1. Only one provider
    logging.verbose(f"{pkgname}: provided by: {', '.join(providers)}")
    if len(providers) == 1:
        return next(iter(providers.values()))

    # 2. Provider with the same package name
    if pkgname in providers:
        logging.verbose(f"{pkgname}: choosing package of the same name as provider")
        return providers[pkgname]

    # 3. Pick a package that will be installed anyway
    for provider_pkgname, provider in providers.items():
        if provider_pkgname in pkgnames_install:
            logging.verbose(
                f"{pkgname}: choosing provider '{provider_pkgname}"
                "', because it will be installed anyway"
            )
            return provider

    # 4. Pick a package that is already installed
    installed = pmb.chroot.apk.installed(suffix)
    for provider_pkgname, provider in providers.items():
        if provider_pkgname in installed:
            logging.verbose(
                f"{pkgname}: choosing provider '{provider_pkgname}"
                f"', because it is installed in the '{suffix}' "
                "chroot already"
            )
            return provider

    # 5. Pick an explicitly selected provider
    provider_pkgname = get_context().config.providers.get(pkgname, "")
    if provider_pkgname in providers:
        logging.verbose(
            f"{pkgname}: choosing provider '{provider_pkgname}', "
            "because it was explicitly selected."
        )
        return providers[provider_pkgname]

    # 6. Pick the provider(s) with the highest priority
    providers = pmb.parse.apkindex.provider_highest_priority(providers, pkgname)
    if len(providers) == 1:
        return next(iter(providers.values()))

    # 7. Pick the shortest provider. (Note: Normally apk would fail here!)
    return pmb.parse.apkindex.provider_shortest(providers, pkgname)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import hashlib
import json
import os
import shutil
import time
import urllib.request
from pathlib import Path
from typing import Any, Literal, overload

import pmb.helpers.cli
import pmb.helpers.run
from pmb.core.context import get_context
from pmb.helpers import logging


def cache_file(prefix: str, url: str) -> Path:
    prefix = prefix.replace("/", "_")
    return Path(f"{prefix}_{hashlib.sha256(url.encode('utf-8')).hexdigest()}")


@overload
def download(
    url: str,
    prefix: str,
    cache: bool = ...,
    loglevel: int = ...,
    allow_404: Literal[False] = ...,
    flush_progress_bar_on_404: bool = ...,
) -> Path: ...


@overload
def download(
    url: str,
    prefix: str,
    cache: bool = ...,
    loglevel: int = ...,
    allow_404: Literal[True] = ...,
    flush_progress_bar_on_404: bool = ...,
) -> Path | None: ...


def download(
    url: str,
    prefix: str,
    cache: bool = True,
    loglevel: int = logging.INFO,
    allow_404: bool = False,
    flush_progress_bar_on_404: bool = False,
) -> Path | None:
    """
    Download a file to disk.

    :param url: the http(s) address of to the file to download
    :param prefix: for the cache, to make it easier to find (cache files
        get a hash of the URL after the prefix)
    :param cache: if True, and url is cached, do not download it again
    :param loglevel: change to logging.DEBUG to only display the download
        message in 'pmbootstrap log', not in stdout.
        We use this when downloading many APKINDEX files at once, no
        point in showing a dozen messages.
    :param allow_404: do not raise an exception when the server responds with a 404 Not Found error.
        Only display a warning on stdout (no matter if loglevel is changed).
    :param flush_progress_bar_on_404: download happens while a progress bar is
        displayed, flush it before printing a warning for 404

    :returns: path to the downloaded file in the cache or None on 404
    """
    # Create cache folder
    context = get_context()
    if not os.path.exists(context.config.work / "cache_http"):
        pmb.helpers.run.user(["mkdir", "-p", context.config.work / "cache_http"])

    # Check if file exists in cache
    path = context.config.work / "cache_http" / cache_file(prefix, url)
    if os.path.exists(path):
        if cache:
            return path
        pmb.helpers.run.user(["rm", path])

    # Offline and not cached
    if context.offline:
        raise RuntimeError(f"File not found in cache and offline flag is enabled: {url}")

    # Download the file
    logging.log(loglevel, "Download " + url)
    try:
        with urllib.request.urlopen(url) as response, open(path, "wb") as handle:
            shutil.copyfileobj(response, handle)
    # Handle 404
    except urllib.error.HTTPError as e:
        if e.code == 404 and allow_404:
            if flush_progress_bar_on_404:
                pmb.helpers.cli.progress_flush()
            logging.warning("WARNING: file not found: " + url)
            return None
        raise

    # Return path in cache
    return path


@overload
def retrieve(
    url: str, headers: dict[str, str] | None = ..., allow_404: Literal[False] = ...
) -> str: ...


@overload
def retrieve(
    url: str, headers: dict[str, str] | None = ..., allow_404: Literal[True] = ...
) -> str | None: ...


def retrieve(
    url: str, headers: dict[str, str] | None = None, allow_404: bool = False
) -> str | None:
    """
    Fetch the content of a URL and returns it as string.

    :param url: the http(s) address of to the resource to fetch
    :param headers: dict of HTTP headers to use
    :param allow_404: do not raise an exception when the server responds with a
        404 Not Found error. Only display a warning

    :returns: str with the content of the response
    """
    # Download the file
    logging.verbose("Retrieving " + url)

    if headers is None:
        headers = {}

    req = urllib.request.Request(url, headers=headers)
    try:
        with urllib.request.urlopen(req) as response:
            return response.read()
    # Handle 404
    except urllib.error.HTTPError as e:
        if e.code == 404 and allow_404:
            logging.warning("WARNING: failed to retrieve content from: " + url)
            return None
        raise


def retrieve_json(url: str, headers: dict[str, str] | None = None) -> Any:
    """
    Fetch the contents of a URL, parse it as JSON and return it.

    See retrieve() for the meaning of the parameters.
    """
    return json.loads(retrieve(url, headers, False))


def measure_latency(url: str) -> float:
    """
    Requests a URL and returns the total time it took to perform the request.

    :param url: the http(s) address of the resource to fetch

    :returns: seconds it took to complete the request
    """
    req = urllib.request.Request(url)
    start_time = time.monotonic()
    with urllib.request.urlopen(req) as response:
        response.read()
    return time.monotonic() - start_time



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
import subprocess
from collections.abc import Sequence
from pathlib import Path, PurePath
from typing import Literal, overload

import pmb.chroot
import pmb.config
import pmb.helpers.run_core
from pmb.core import Chroot
from pmb.types import (
    Env,
    PathString,
    RunOutputType,
    RunOutputTypeDefault,
    RunOutputTypePopen,
    RunReturnType,
)


def rootm(
    cmds: Sequence[Sequence[PathString]],
    chroot: Chroot = Chroot.native(),
    working_dir: PurePath = PurePath("/"),
    output: RunOutputType = RunOutputTypeDefault.LOG,
    output_return: bool = False,
    check: bool | None = None,
    env: Env = {},
    disable_timeout: bool = False,
    add_proxy_env_vars: bool = True,
) -> RunReturnType:
    """
    Run a list of commands inside a chroot as root.

    :param env: dict of environment variables to be passed to the command, e.g.
                {"JOBS": "5"}
    :param working_dir: chroot-relative working directory
    :param add_proxy_env_vars: if True, preserve HTTP_PROXY etc. vars from host
                               environment. pmb.chroot.user sets this to False
                               when calling pmb.chroot.root, because it already
                               makes the variables part of the cmd argument.

    See pmb.helpers.run_core.core() for a detailed description of all other
    arguments and the return value.
    """
    # Convert any Path objects to their string representation
    cmd_strs = [[os.fspath(x) for x in cmd] for cmd in cmds]

    # Readable log message (without all the escaping)
    msg = f"({chroot}) % "
    for key, value in env.items():
        msg += f"{key}={value} "
    if working_dir != PurePath("/"):
        msg += f"cd {working_dir}; "
    msg += "; ".join([" ".join(cmd_str) for cmd_str in cmd_strs])

    # Merge env with defaults into env_all
    env_all: Env = {
        "CHARSET": "UTF-8",
        "HISTFILE": "~/.ash_history",
        "HOME": "/root",
        "LANG": "UTF-8",
        "PATH": pmb.config.chroot_path,
        "PYTHONUNBUFFERED": "1",
        "SHELL": "/bin/ash",
        "TERM": "xterm",
    }
    for key, value in env.items():
        env_all[key] = value
    if add_proxy_env_vars:
        pmb.helpers.run_core.add_proxy_env_vars(env_all)

    # Build the command in steps and run it, e.g.:
    # cmd: ["echo", "test"]
    # cmd_chroot: ["/sbin/chroot", "/..._native", "/bin/sh", "-c", "echo test"]
    # cmd_sudo: ["sudo", "env", "-i", "sh", "-c", "PATH=... /sbin/chroot ..."]
    executables = pmb.config.required_programs
    cmd_chroot: list[PathString] = [
        executables["chroot"],
        chroot.path,
        "/bin/sh",
        "-c",
        pmb.helpers.run_core.flat_cmd(cmd_strs, Path(working_dir)),
    ]
    cmd_sudo = pmb.config.sudo(
        [
            "env",
            "-i",
            executables["sh"],
            "-c",
            pmb.helpers.run_core.flat_cmd([cmd_chroot], env=env_all),
        ]
    )
    return pmb.helpers.run_core.core(
        msg, cmd_sudo, None, output, output_return, check, True, disable_timeout
    )


@overload
def root(
    cmds: Sequence[PathString],
    chroot: Chroot = ...,
    *,
    check: bool = ...,
) -> int: ...


@overload
def root(
    cmds: Sequence[PathString],
    chroot: Chroot = ...,
    working_dir: PurePath = ...,
    output: RunOutputTypePopen = ...,
    output_return: Literal[False] = ...,
    check: bool | None = ...,
    env: Env = ...,
    disable_timeout: bool = ...,
    add_proxy_env_vars: bool = ...,
) -> subprocess.Popen: ...


@overload
def root(
    cmds: Sequence[PathString],
    chroot: Chroot = ...,
    working_dir: PurePath = ...,
    output: RunOutputTypeDefault = ...,
    output_return: Literal[False] = ...,
    check: bool | None = ...,
    env: Env = ...,
    disable_timeout: bool = ...,
    add_proxy_env_vars: bool = ...,
) -> int: ...


@overload
def root(
    cmds: Sequence[PathString],
    chroot: Chroot = ...,
    working_dir: PurePath = ...,
    output: RunOutputType = ...,
    output_return: Literal[True] = ...,
    check: bool | None = ...,
    env: Env = ...,
    disable_timeout: bool = ...,
    add_proxy_env_vars: bool = ...,
) -> str: ...


@overload
def root(
    cmds: Sequence[PathString],
    chroot: Chroot = ...,
    working_dir: PurePath = ...,
    output: RunOutputType = ...,
    output_return: bool = ...,
    check: bool | None = ...,
    env: Env = ...,
    disable_timeout: bool = ...,
    add_proxy_env_vars: bool = ...,
) -> subprocess.Popen | int | str: ...


def root(
    cmds: Sequence[PathString],
    chroot: Chroot = Chroot.native(),
    working_dir: PurePath = PurePath("/"),
    output: RunOutputType = RunOutputTypeDefault.LOG,
    output_return: bool = False,
    check: bool | None = None,
    env: Env = {},
    disable_timeout: bool = False,
    add_proxy_env_vars: bool = True,
) -> RunReturnType:
    return rootm(
        [cmds],
        chroot,
        working_dir,
        output,
        output_return,
        check,
        env,
        disable_timeout,
        add_proxy_env_vars,
    )


def userm(
    cmds: Sequence[Sequence[PathString]],
    chroot: Chroot = Chroot.native(),
    working_dir: Path = Path("/"),
    output: RunOutputType = RunOutputTypeDefault.LOG,
    output_return: bool = False,
    check: bool | None = None,
    env: Env = {},
) -> RunReturnType:
    """
    Run a command inside a chroot as "user". We always use the BusyBox
    implementation of 'su', because other implementations may override the PATH
    environment variable (#1071).

    :param env: dict of environment variables to be passed to the command, e.g.
                {"JOBS": "5"}

    See pmb.helpers.run_core.core() for a detailed description of all other
    arguments and the return value.
    """
    env = env.copy()
    pmb.helpers.run_core.add_proxy_env_vars(env)

    if "HOME" not in env:
        env["HOME"] = "/home/pmos"

    flat_cmd = pmb.helpers.run_core.flat_cmd(cmds, env=env)
    cmd = ["busybox", "su", "pmos", "-c", flat_cmd]
    return pmb.chroot.root(
        cmd, chroot, working_dir, output, output_return, check, {}, add_proxy_env_vars=False
    )


@overload
def user(
    cmd: Sequence[PathString],
    chroot: Chroot = ...,
    working_dir: Path = ...,
    output: RunOutputTypePopen = ...,
    output_return: Literal[False] = ...,
    check: bool | None = ...,
    env: Env = ...,
) -> subprocess.Popen: ...


@overload
def user(
    cmd: Sequence[PathString],
    chroot: Chroot = ...,
    working_dir: Path = ...,
    output: RunOutputTypeDefault = ...,
    output_return: Literal[False] = ...,
    check: bool | None = ...,
    env: Env = ...,
) -> int: ...


@overload
def user(
    cmd: Sequence[PathString],
    chroot: Chroot = ...,
    working_dir: Path = ...,
    output: RunOutputType = ...,
    output_return: Literal[True] = ...,
    check: bool | None = ...,
    env: Env = ...,
) -> str: ...


def user(
    cmd: Sequence[PathString],
    chroot: Chroot = Chroot.native(),
    working_dir: Path = Path("/"),
    output: RunOutputType = RunOutputTypeDefault.LOG,
    output_return: bool = False,
    check: bool | None = None,
    env: Env = {},
) -> RunReturnType:
    return userm([cmd], chroot, working_dir, output, output_return, check, env)


def user_exists(username: str, chroot: Chroot = Chroot.native()) -> bool:
    """
    Checks if username exists in the system

    :param username: User name
    :returns: bool
    """
    output = pmb.chroot.root(
        ["getent", "passwd", username], chroot, output_return=True, check=False
    )
    return len(output) > 0



# Copyright 2025 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import fnmatch
import logging

import pmb.aportgen.core
import pmb.helpers.git
import pmb.helpers.run
from pmb.core.arch import Arch
from pmb.core.context import get_context
from pmb.core.pkgrepo import pkgrepo_default_path


def depends_for_sonames(libraries: dict[str, str], arch_libc: Arch) -> list:
    """
    Get packages providing specific sonames from Alpine's main repo. Usually
    this would be done during package builds using abuild's "tracedeps". But
    this leads to our cross gccs immediately breaking once Alpine gcc packages
    change. So we figure out depends on our own here.
    """
    apkindex_main = pmb.helpers.repo.apkindex_files(
        arch_libc, user_repository=False, exclude_mirrors=["pmaports", "systemd"]
    )[0]
    apkindex = pmb.parse.apkindex.parse(apkindex_main, True)

    result: dict[str, str] = {}
    for pattern_soname in libraries:
        pattern_pkgname = libraries[pattern_soname]

        for provide in apkindex:
            if not fnmatch.fnmatch(provide, pattern_soname):
                continue
            match = None
            for pkgname in apkindex[provide]:
                if fnmatch.fnmatch(pkgname, pattern_pkgname):
                    logging.info(f"{provide}: provided by {pkgname}")
                    match = pkgname
                    # No break, so it prints other matches too if any. This
                    # should make debugging easier if something goes wrong.
                else:
                    logging.warning(
                        f"{provide}: provided by {pkgname} which is an unexpected pkgname, ignoring..."
                    )
            if match:
                if pattern_soname in result:
                    old = result[pattern_soname].split(".so.")[1]
                    new = provide.split(".so.")[1]
                    if pmb.parse.version.compare(new, old) == 1:
                        logging.debug(
                            f"{provide}: new highest version found for pattern {pattern_soname}"
                        )
                        result[pattern_soname] = provide
                else:
                    logging.debug(f"{provide}: first version found for pattern {pattern_soname}")
                    result[pattern_soname] = provide

        if pattern_soname not in result:
            raise RuntimeError(
                f"{pattern_soname}: is not provided by any package, can't trace dependencies for this pattern."
            )

    return list(result.values())


def generate(pkgname: str) -> None:
    # Copy original aport
    prefix = pkgname.split("-")[0]
    arch = Arch.from_str(pkgname.split("-")[1])
    # Until pmb#2517 is resolved properly, we set the tracedeps manually. The
    # musl soname contains the architecture name, support cross compiling from
    # aarch64 to x86_64 and x86_64 to all other arches.
    arch_libc = Arch.from_str("aarch64" if pkgname.split("-")[1] == "x86_64" else "x86_64")
    context = get_context()
    if prefix == "gcc":
        upstream = pmb.aportgen.core.get_upstream_aport("gcc", arch)
        based_on = "main/gcc (from Alpine)"
    elif prefix == "gcc4":
        upstream = pkgrepo_default_path() / "main/gcc4"
        based_on = "main/gcc4 (from postmarketOS)"
    elif prefix == "gcc6":
        upstream = pkgrepo_default_path() / "main/gcc6"
        based_on = "main/gcc6 (from postmarketOS)"
    else:
        raise ValueError(f"Invalid prefix '{prefix}', expected gcc, gcc4 or gcc6.")
    pmb.helpers.run.user(["cp", "-r", upstream, context.config.work / "aportgen"])

    # Rewrite APKBUILD
    fields = {
        "pkgname": pkgname,
        "pkgdesc": f"Stage2 cross-compiler for {arch}",
        "arch": pmb.aportgen.get_cross_package_arches(pkgname),
        "depends": f"binutils-{arch} mpc1",
        "makedepends_build": "gcc g++ bison flex texinfo gawk zip"
        " gmp-dev mpfr-dev mpc1-dev zlib-dev",
        "makedepends_host": "linux-headers gmp-dev mpfr-dev mpc1-dev isl-dev"
        f" zlib-dev musl-dev-{arch} binutils-{arch}",
        "subpackages": "",
        # gcc6: options is already there, so we need to replace it and not only
        # set it below the header like done below.
        # !tracedeps: workaround for issue 2517
        "options": "!strip !tracedeps toolchain",
        "LIBGOMP": "false",
        "LIBGCC": "false",
        "LIBATOMIC": "false",
        "LIBITM": "false",
    }

    libraries = {
        f"so:libc.musl-{arch_libc}.so.*": "musl",
        "so:libgcc_s.so.*": "libgcc",
        "so:libgmp.so.*": "gmp",
        "so:libisl.so.*": "isl*",
        "so:libmpc.so.*": "mpc1",
        "so:libmpfr.so.*": "mpfr4",
        "so:libstdc++.so.*": "libstdc++",
        "so:libz.so.*": "zlib",
    }
    logging.info(f"*** Getting sonames for depends (arch_libc: {arch_libc})")
    fields["depends"] += f" {' '.join(depends_for_sonames(libraries, arch_libc))}"

    # Latest gcc only, not gcc4 and gcc6
    if prefix == "gcc":
        fields["subpackages"] = f"g++-{arch}:gpp libstdc++-dev-{arch}:libcxx_dev"

    below_header = (
        "CTARGET_ARCH="
        + str(arch)
        + """
        CTARGET="$(arch_to_hostspec ${CTARGET_ARCH})"
        LANG_D=false
        LANG_OBJC=false
        LANG_JAVA=false
        LANG_GO=false
        LANG_FORTRAN=false
        LANG_ADA=false
        options="!strip"

        # abuild doesn't try to tries to install "build-base-$CTARGET_ARCH"
        # when this variable matches "no*"
        BOOTSTRAP="nobuildbase"

        # abuild will only cross compile when this variable is set, but it
        # needs to find a valid package database in there for dependency
        # resolving, so we set it to /.
        CBUILDROOT="/"

        _cross_configure="--disable-bootstrap --with-sysroot=/usr/$CTARGET"
    """
    )

    libraries_gpp = {
        f"so:libc.musl-{arch_libc}.so.*": "musl",
        "so:libgmp.so.*": "gmp",
        "so:libisl.so.*": "isl*",
        "so:libmpc.so.*": "mpc1",
        "so:libmpfr.so.*": "mpfr4",
        "so:libz.so.*": "zlib",
    }
    logging.info(f"*** Getting sonames for depends in gpp subpackage (arch_libc: {arch_libc})")
    depends_gpp = " ".join(depends_for_sonames(libraries_gpp, arch_libc))

    replace_simple = {
        # Do not package libstdc++, do not add "g++-$ARCH" here (already
        # did that explicitly in the subpackages variable above, so
        # pmbootstrap picks it up properly).
        '*subpackages="$subpackages libstdc++:libcxx:*': None,
        # We set the cross_configure variable at the beginning, so it does not
        # use CBUILDROOT as sysroot. In the original APKBUILD this is a local
        # variable, but we make it a global one.
        "*_cross_configure=*": None,
        # Do not build foreign arch libgcc, we use the one from Alpine (#2168)
        "_libgcc=true*": "_libgcc=false",
        # Keep the cross prefix in package()
        "*# These are moved into packages with arch=*": "",
        "*# cross prefix (doesn't exist when BOOTSTRAP=nolibc)*": "",
        '*BOOTSTRAP" != nolibc ] && mv *': "",
        # Add depends to the gpp subpackage, so we don't need to use tracedeps
        "*amove $_gcclibexec/cc1plus*": f'\tdepends="$depends {depends_gpp}"\n\n\tamove $_gcclibexec/cc1plus',
        # Disable libsanitizer for e.g. aarch64 -> x86_64 gcc (pma!6722)
        '*_sanitizer_configure="--enable-libsanitizer"*': "",
    }

    pmb.aportgen.core.rewrite(
        pkgname, based_on, fields, replace_simple=replace_simple, below_header=below_header
    )



# Copyright 2024 Stefan Hansson
# SPDX-License-Identifier: GPL-3.0-or-later
from dataclasses import asdict, dataclass

from pmb.core.arch import Arch


@dataclass
class ApkindexBlock:
    """
    "timestamp" and "origin" are not set for virtual packages (#1273).
    We use that information to skip these virtual packages in parse().
    """

    #: the architecture of the package
    arch: Arch
    #: dependencies for the package
    depends: list[str]
    #: the origin name of the package
    origin: str | None
    #: package name
    pkgname: str
    #: what this package provides
    provides: list[str]
    #: provider priority for the package
    provider_priority: int | None
    #: unix timestamp of the package build date/time
    timestamp: str | None
    #: package version
    version: str

    @property
    def __dict__(self) -> dict:
        """This needs a manual implementation as the dataclass contains an enum."""
        block_dict = asdict(self)

        block_dict["arch"] = str(block_dict["arch"])

        return block_dict

    @__dict__.setter
    def __dict__(self, new_value: dict) -> None:
        raise AssertionError("Use dot operator access for ApkindexBlock")



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import configparser
import os
import sys
from pathlib import Path
from typing import Final

import pmb.config
import pmb.helpers.git
import pmb.helpers.pmaports
import pmb.parse.version
from pmb.core.pkgrepo import (
    pkgrepo_default_path,
    pkgrepo_name,
    pkgrepo_paths,
    pkgrepo_relative_path,
)
from pmb.helpers import logging
from pmb.helpers.exceptions import NonBugError
from pmb.meta import Cache
from pmb.types import RunOutputTypeDefault


def clone(do_shallow: bool = False) -> None:
    logging.info("Setting up the native chroot and cloning the package build recipes (pmaports)...")

    # Set up the native chroot and clone pmaports
    pmb.helpers.git.clone("pmaports", do_shallow)


def check_version_pmaports(real: str) -> None:
    # Compare versions
    min = pmb.config.pmaports_min_version
    if pmb.parse.version.compare(real, min) >= 0:
        return

    # Outated error
    logging.info(
        "NOTE: your pmaports folder has version "
        + real
        + ", but"
        + " version "
        + min
        + " is required."
    )
    raise RuntimeError("Run 'pmbootstrap pull' to update your pmaports.")


def check_version_pmbootstrap(min_ver: str) -> None:
    # Compare versions
    real = pmb.__version__
    if pmb.parse.version.compare(real, min_ver) >= 0:
        return

    # Show versions
    logging.info(
        f"NOTE: you are using pmbootstrap version {real}, but version {min_ver} is required."
    )

    # Error for git clone
    pmb_src = pmb.config.pmb_src
    if (pmb_src / ".git").exists():
        raise RuntimeError(
            "Please update your local pmbootstrap repository."
            f" Usually with: 'git -C \"{pmb_src}\" pull'"
        )

    # Error for package manager installation
    raise RuntimeError(
        "Please update your pmbootstrap version (with your"
        " distribution's package manager, or with pip, "
        " depending on how you have installed it). If that is"
        " not possible, consider cloning the latest version"
        " of pmbootstrap from git."
    )


@Cache()
def read_config_repos() -> dict[str, configparser.SectionProxy]:
    """Read the sections starting with "repo:" from pmaports.cfg."""
    cfg = configparser.ConfigParser()
    cfg.read(f"{pkgrepo_default_path()}/pmaports.cfg")

    ret = {}
    for section in cfg:
        if not section.startswith("repo:"):
            continue
        repo = section.split("repo:", 1)[1]
        ret[repo] = cfg[section]

    return ret


@Cache("aports", "add_systemd_prefix")
def read_config(
    aports: Path | None = None, add_systemd_prefix: bool = True
) -> configparser.SectionProxy:
    """
    Read and verify pmaports.cfg. If aports is not
    specified and systemd is enabled, the returned channel
    will be the systemd one (e.g. systemd-edge instead of edge)
    since we'll use the first pkgrepo which is systemd.
    """
    if aports is None:
        aports = pkgrepo_paths()[0]

    systemd = pkgrepo_name(aports) == "systemd"
    # extra-repos don't have a pmaports.cfg
    # so jump up the main aports dir
    if "extra-repos" in aports.parts:
        aports = pkgrepo_relative_path(aports)[0]

    # Migration message
    if not os.path.exists(aports):
        logging.error(f"ERROR: pmaports dir not found: {aports}")
        logging.error("Did you run 'pmbootstrap init'?")
        sys.exit(1)

    # Require the config
    path_cfg = aports / "pmaports.cfg"
    if not os.path.exists(path_cfg):
        raise RuntimeError(f"Invalid pmaports repository, could not find the config: {path_cfg}")

    # Load the config
    cfg = configparser.ConfigParser()
    cfg.read(path_cfg)
    ret = cfg["pmaports"]

    # Version checks
    check_version_pmaports(ret["version"])
    check_version_pmbootstrap(ret["pmbootstrap_min_version"])

    # Translate legacy channel names
    ret["channel"] = pmb.helpers.pmaports.get_channel_new(ret["channel"])

    if systemd and add_systemd_prefix:
        ret["channel"] = "systemd-" + ret["channel"]

    return ret


def all_channels() -> list[str]:
    """Get a list of all channels for all pkgrepos."""
    ret = {read_config(repo)["channel"] for repo in pkgrepo_paths()}

    logging.verbose(f"all_channels: {ret}")
    return list(ret)


def read_config_channel() -> dict[str, str]:
    """
    Get the properties of the currently active channel in pmaports.git.

    As specified in channels.cfg (https://postmarketos.org/channels.cfg).

    :returns: {"description: ...,
               "branch_pmaports": ...,
               "branch_aports": ...,
               "mirrordir_alpine": ...}

    """
    aports = pkgrepo_default_path()
    channel = read_config(aports)["channel"]
    channels_cfg = pmb.helpers.git.parse_channels_cfg(aports)

    if channel in channels_cfg["channels"]:
        return channels_cfg["channels"][channel]

    # Channel not in channels.cfg, try to be helpful
    branch = pmb.helpers.git.rev_parse(aports, extra_args=["--abbrev-ref"])
    remote = pmb.helpers.git.get_upstream_remote(aports)
    logging.info(
        "NOTE: fix the error by rebasing or cherry picking relevant"
        " commits from this branch onto a branch that is on a"
        " supported channel: main, v24.06, …"
    )
    logging.info(
        "NOTE: as workaround, you may pass --config-channels with a"
        " custom channels.cfg. Reference:"
        " https://postmarketos.org/channels.cfg"
    )
    raise RuntimeError(
        f"Current branch '{branch}' of pmaports.git is on"
        f" channel '{channel}', but this channel was not"
        f" found in channels.cfg (of {remote}/main"
        " branch). Looks like a very old branch."
    )


def init(do_shallow_clone: bool = False) -> None:
    if not os.path.exists(pkgrepo_default_path()):
        clone(do_shallow_clone)
    read_config()


DEVELOPMENT_CHANNEL: Final[str] = "edge"


def switch_to_channel_branch(channel_new: str) -> bool:
    """
    Checkout the channel's branch in pmaports.git.

    :channel_new: channel name (e.g. "edge", "v21.03")

    :returns: True if another branch was checked out, False otherwise
    """
    aports = pkgrepo_default_path()

    # list current and new branches/channels
    channels_cfg = pmb.helpers.git.parse_channels_cfg(aports)
    branch_new = channels_cfg["channels"][channel_new]["branch_pmaports"]
    branch_current = pmb.helpers.git.rev_parse(aports, extra_args=["--abbrev-ref"])

    # Check current pmaports branch channel
    channel_current = read_config(aports)["channel"]
    # If we are on edge and on master but edge calls for main
    if (
        channel_current == DEVELOPMENT_CHANNEL
        and branch_current == "master"
        and branch_new == "main"
    ):
        logging.info("NOTE: master was renamed to main, switching to it")
        branch_new = "main"
    elif channel_current == channel_new:
        return False

    if (
        branch_current == "master_staging_systemd"
        and channel_new == DEVELOPMENT_CHANNEL
        and pmb.config.is_systemd_selected()
    ):
        logging.info("NOTE: master_staging_systemd was merged into main, switching to it")

    logging.info(
        f"Currently checked out branch '{branch_current}' of"
        f" pmaports.git is on channel '{channel_current}'."
    )
    logging.info(f"Switching to branch '{branch_new}' on channel '{channel_new}'...")

    # Make sure we don't have mounts related to the old channel
    pmb.chroot.shutdown()

    # We need to make sure the branch exists in case --shallow-initial-clone previously was used.
    # Checking whether args.shallow_initial_clone is True would not solve this problem as a user
    # could use --shallow-initial-clone when cloning for the first time, then call init without the
    # flag subsequent times without the repository ever ending up "deep cloned". As such, check for
    # the presence of the branch instead of whether the flag is True.
    if pmb.helpers.run.user(
        ["git", "show-ref", "--quiet", "--branches", branch_new], aports, check=False
    ):
        logging.info(f"Branch {branch_new} doesn't exist, attempting to fetch it")

        if pmb.helpers.run.user(
            ["git", "fetch", "--depth=1", "origin", f"{branch_new}:{branch_new}"],
            aports,
            RunOutputTypeDefault.INTERACTIVE,
            check=False,
        ):
            raise NonBugError(f"Failed to fetch {branch_new}!")

    # Attempt to switch branch (git gives a nice error message, mentioning
    # which files need to be committed/stashed, so just pass it through)
    if pmb.helpers.run.user(
        ["git", "switch", branch_new], aports, RunOutputTypeDefault.INTERACTIVE, check=False
    ):
        raise RuntimeError(
            "Failed to switch branch. Go to your pmaports and"
            " fix what git complained about, then try again: "
            f"{aports}"
        )

    # Verify pmaports.cfg on new branch
    read_config()
    return True


def install_githooks() -> None:
    aports = pkgrepo_default_path()
    hooks_dir = aports / ".githooks"
    if not hooks_dir.exists():
        logging.info("No .githooks dir found")
        return
    for h in os.listdir(hooks_dir):
        src = os.path.join(hooks_dir, h)
        # Use git default hooks dir so users can ignore our hooks
        # if they dislike them by setting "core.hooksPath" git config
        dst = aports / ".git/hooks" / h
        if pmb.helpers.run.user(["cp", src, dst], check=False):
            logging.warning(f"WARNING: Copying git hook failed: {dst}")



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import copy
import inspect
import os
from dataclasses import dataclass
from enum import Enum
from pathlib import Path

import pmb.config
import pmb.helpers.devices
from pmb.core.arch import Arch
from pmb.core.context import get_context
from pmb.helpers import logging
from pmb.helpers.exceptions import NonBugError
from pmb.meta import Cache


class InitfsCompressionFormat(Enum):
    ZSTD = "zstd"
    LZ4 = "lz4"
    LZMA = "lzma"
    GZIP = "gzip"
    NONE = "none"

    @staticmethod
    def from_str(compression_format: str) -> InitfsCompressionFormat:
        try:
            return InitfsCompressionFormat(compression_format)
        except ValueError as exception:
            raise ValueError(f"Invalid compression format '{compression_format}'") from exception


class InitfsCompressionLevel(Enum):
    DEFAULT = "default"
    FAST = "fast"
    BEST = "best"

    @staticmethod
    def from_str(compression_level: str) -> InitfsCompressionLevel:
        try:
            return InitfsCompressionLevel(compression_level)
        except ValueError as exception:
            raise ValueError(f"Invalid compression level '{compression_level}'") from exception


@dataclass
class InitfsCompression:
    format_: InitfsCompressionFormat
    level: InitfsCompressionLevel | None

    @staticmethod
    def from_str(initfs_compression: str) -> InitfsCompression:
        segments = initfs_compression.split(":", maxsplit=1)

        # If we can't even figure out the format, the other information is no use. As such, let this
        # raise an exception if it happens.
        format_ = InitfsCompressionFormat.from_str(segments[0])

        try:
            level = InitfsCompressionLevel.from_str(segments[1]) if len(segments) == 2 else None
        except ValueError:
            level = None

        return InitfsCompression(format_, level)


# FIXME: It feels weird to handle this at parse time.
# we should instead have the Deviceinfo object store
# the attributes for all kernels and require the user
# to specify which one they're using.
# Basically: treat Deviceinfo as a standalone type that
# doesn't need to traverse pmaports.
def _parse_kernel_suffix(info: dict[str, str], device: str, kernel: str | None) -> dict[str, str]:
    """
    Remove the kernel suffix (as selected in 'pmbootstrap init') from
    deviceinfo variables. Related:
    https://wiki.postmarketos.org/wiki/Device_specific_package#Multiple_kernels

    :param info: deviceinfo dict, e.g.:
                 {"a": "first",
                  "b_mainline": "second",
                  "b_downstream": "third"}
    :param device: which device info belongs to
    :param kernel: which kernel suffix to remove (e.g. "mainline")
    :returns: info, but with the configured kernel suffix removed, e.g:
              {"a": "first",
               "b": "second",
               "b_downstream": "third"}
    """
    # We don't support parsing the kernel variants in tests yet, since this code
    # depends on pmaports being available and calls into a whole lot of other code.
    if os.environ.get("PYTEST_CURRENT_TEST", "").startswith("pmb/parse/test_deviceinfo.py"):
        # If you hit this, you're probably trying to add a test for kernel variants.
        # You'll need to figure out how to mock the APKBUILD parsing below.
        assert kernel is None
        return info
    # Do nothing if the configured kernel isn't available in the kernel (e.g.
    # after switching from device with multiple kernels to device with only one
    # kernel)
    kernels = pmb.parse._apkbuild.kernels(device)
    if not kernels or kernel not in kernels:
        logging.verbose(f"parse_kernel_suffix: {kernel} not in {kernels}")
        return info

    ret = copy.copy(info)

    suffix_kernel = kernel.replace("-", "_")
    for key in inspect.get_annotations(Deviceinfo):
        key_kernel = f"{key}_{suffix_kernel}"
        if key_kernel not in ret:
            continue

        # Move ret[key_kernel] to ret[key]
        logging.verbose(f"parse_kernel_suffix: {key_kernel} => {key}")
        ret[key] = ret[key_kernel]
        del ret[key_kernel]

    return ret


@Cache("device", "kernel")
def deviceinfo(device: str | None = None, kernel: str | None = None) -> Deviceinfo:
    """
    :param device: defaults to args.device
    :param kernel: defaults to args.kernel
    """
    context = get_context()
    if not device:
        device = context.config.device
    if not kernel:
        kernel = context.config.kernel

    path = pmb.helpers.devices.find_path(device, "deviceinfo")
    if not path:
        raise NonBugError(
            f"Device '{device}' not found. Run 'pmbootstrap init' to start a new device port or to "
            "choose another device. It may have been renamed, see <https://postmarketos.org/renamed>"
        )

    return Deviceinfo(path, kernel)


class Deviceinfo:
    """Variables from deviceinfo. Reference: <https://postmarketos.org/deviceinfo>"""

    path: Path
    # general
    format_version: str
    name: str
    manufacturer: str
    codename: str
    year: str
    dtb: str = ""
    arch: Arch

    # device
    chassis: str | None
    keyboard: str | None = ""  # deprecated
    drm: bool | None = False
    dev_touchscreen: str | None = ""
    dev_touchscreen_calibration: str | None = ""
    append_dtb: str | None = ""

    # bootloader
    flash_method: str = ""
    boot_filesystem: str | None = ""
    create_initfs_extra: bool | None = False
    create_prep_boot: bool | None = False
    initfs_compression: InitfsCompression = InitfsCompression(InitfsCompressionFormat.GZIP, None)

    # flash
    flash_heimdall_partition_kernel: str | None = ""
    flash_heimdall_partition_vendor_boot: str | None = ""
    flash_heimdall_partition_initfs: str | None = ""
    flash_heimdall_partition_rootfs: str | None = ""
    flash_heimdall_partition_system: str | None = ""  # deprecated
    flash_heimdall_partition_vbmeta: str | None = ""
    flash_heimdall_partition_dtbo: str | None = ""
    flash_fastboot_partition_kernel: str | None = ""
    flash_fastboot_partition_vendor_boot: str | None = ""
    flash_fastboot_partition_rootfs: str | None = ""
    flash_fastboot_partition_system: str | None = ""  # deprecated
    flash_fastboot_partition_vbmeta: str | None = ""
    flash_fastboot_partition_dtbo: str | None = ""
    flash_rk_partition_kernel: str | None = ""
    flash_rk_partition_vendor_boot: str | None = ""
    flash_rk_partition_rootfs: str | None = ""
    flash_rk_partition_system: str | None = ""  # deprecated
    flash_mtkclient_partition_kernel: str | None = ""
    flash_mtkclient_partition_vendor_boot: str | None = ""
    flash_mtkclient_partition_rootfs: str | None = ""
    flash_mtkclient_partition_vbmeta: str | None = ""
    flash_mtkclient_partition_dtbo: str | None = ""
    generate_legacy_uboot_initfs: str | None = ""
    kernel_cmdline: str | None = ""
    generate_bootimg: str | None = ""
    header_version: int | None = None
    bootimg_qcdt: str | None = ""
    bootimg_qcdt_type: str | None = ""
    bootimg_mtk_mkimage: str | None = ""  # deprecated
    bootimg_mtk_label_kernel: str | None = ""
    bootimg_mtk_label_ramdisk: str | None = ""
    bootimg_dtb_second: str | None = ""
    bootimg_custom_args: str | None = ""
    flash_offset_base: str | None = ""
    flash_offset_dtb: str | None = ""
    flash_offset_kernel: str | None = ""
    flash_offset_ramdisk: str | None = ""
    flash_offset_second: str | None = ""
    flash_offset_tags: str | None = ""
    flash_pagesize: str | None = ""
    flash_fastboot_max_size: str | None = ""
    flash_sparse: str | None = ""
    flash_sparse_samsung_format: str | None = ""
    rootfs_image_sector_size: str | None = ""
    sd_embed_firmware: str | None = ""
    sd_embed_firmware_step_size: str | None = ""
    partition_blacklist: str | None = ""
    boot_part_start: str | None = ""
    partition_type: str | None = ""
    root_filesystem: str | None = ""
    flash_kernel_on_update: str | None = ""
    cgpt_kpart: str | None = ""
    cgpt_kpart_start: str | None = ""
    cgpt_kpart_size: str | None = ""

    # weston
    weston_pixman_type: str | None = ""

    # keymaps
    keymaps: str | None = ""

    def __init__(self, path: Path, kernel: str | None = None) -> None:
        ret = {}
        with open(path) as handle:
            for line in handle:
                if not line.startswith("deviceinfo_"):
                    continue
                if "=" not in line:
                    raise SyntaxError(f"{path}: No '=' found:\n\t{line}")
                split = line.split("=", 1)
                key = split[0][len("deviceinfo_") :]
                value = split[1].replace('"', "").replace("\n", "")
                ret[key] = value

        ret = _parse_kernel_suffix(ret, ret["codename"], kernel)

        for key, value in ret.items():
            # FIXME: something to turn on and fix in the future
            # Should be rewritten to use deviceinfo_schema.toml as source of truth instead of the
            # Deviceinfo class keys first though.
            # if key not in Deviceinfo.__annotations__.keys():
            #     logging.warning(f"deviceinfo: {key} is not a known attribute")
            match key:
                case "arch":
                    self.arch = Arch.from_str(value)
                case "gpu_accelerated":  # deprecated
                    self.drm = value == "true"
                case "header_version":
                    self.header_version = int(value)
                case "initfs_compression":
                    self.initfs_compression = InitfsCompression.from_str(value)
                case _:
                    setattr(self, key, value)

        if not self.flash_method:
            self.flash_method = "none"



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
from pathlib import Path, PurePath

import pmb.helpers
import pmb.helpers.run
from pmb.core import Chroot
from pmb.types import PathString


def ismount(folder: Path) -> bool:
    """
    Ismount() implementation that works for mount --bind.

    Workaround for: https://bugs.python.org/issue29707
    """
    folder = folder.resolve()
    with open("/proc/mounts") as handle:
        for line in handle:
            words = line.split()
            if len(words) >= 2 and Path(words[1]) == folder:
                return True
            if words[0] == str(folder):
                return True
    return False


def bind(
    source: PathString, destination: Path, create_folders: bool = True, umount: bool = False
) -> None:
    """
    Mount --bind a folder and create necessary directory structure.

    :param umount: when destination is already a mount point, umount it first.
    """
    # Check/umount destination
    if ismount(destination):
        if umount:
            umount_all(destination)
        else:
            return

    # Check/create folders
    for path in [source, destination]:
        if os.path.exists(path):
            continue
        if create_folders:
            pmb.helpers.run.root(["mkdir", "-p", path])
        else:
            raise RuntimeError(f"Mount failed, folder does not exist: {path}")

    # Actually mount the folder
    pmb.helpers.run.root(["mount", "--bind", source, destination])

    # Verify that it has worked
    if not ismount(destination):
        raise RuntimeError(f"Mount failed: {source} -> {destination}")


def bind_file(source: Path, destination: Path, create_folders: bool = False) -> None:
    """Mount a file with the --bind option, and create the destination file, if necessary."""
    # Skip existing mountpoint
    if ismount(destination):
        return

    # Create empty file
    if not destination.exists():
        if create_folders:
            dest_dir: Path = destination.parent
            if not dest_dir.is_dir():
                pmb.helpers.run.root(["mkdir", "-p", dest_dir])

        pmb.helpers.run.root(["touch", destination])

    # Mount
    pmb.helpers.run.root(["mount", "--bind", source, destination])


def umount_all_list(prefix: Path, source: Path = Path("/proc/mounts")) -> list[Path]:
    """
    Parse `/proc/mounts` for all folders beginning with a prefix.

    :source: can be changed for testcases

    :returns: a list of folders that need to be umounted

    """
    ret = []
    prefix = prefix.resolve()
    with source.open() as handle:
        for line in handle:
            words = line.split()
            if len(words) < 2:
                raise RuntimeError(f"Failed to parse line in {source}: {line}")
            mountpoint = Path(words[1].replace(r"\040(deleted)", ""))
            if mountpoint.is_relative_to(prefix):  # is subpath
                ret.append(mountpoint)
    ret.sort(reverse=True)
    return ret


def umount_all(folder: Path) -> None:
    """Umount all folders that are mounted inside a given folder."""
    for mountpoint in umount_all_list(folder):
        pmb.helpers.run.root(["umount", mountpoint])
        if ismount(mountpoint):
            raise RuntimeError(f"Failed to umount: {mountpoint}")


def mount_device_rootfs(chroot_rootfs: Chroot, chroot_base: Chroot = Chroot.native()) -> PurePath:
    """
    Mount the device rootfs.

    :param chroot_rootfs: the chroot where the rootfs that will be
                          installed on the device has been created (e.g.
                          "rootfs_qemu-amd64")
    :param chroot_base: the chroot rootfs mounted to
    :returns: the mountpoint (relative to the chroot)
    """
    mountpoint = PurePath("/mnt", str(chroot_rootfs))
    pmb.helpers.mount.bind(chroot_rootfs.path, chroot_base / mountpoint)
    return mountpoint



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import pmb.chroot.apk
import pmb.config
import pmb.config.pmaports
import pmb.helpers.mount
from pmb.core import Chroot, ChrootType
from pmb.helpers.mount import mount_device_rootfs


def install_depends(method: str) -> None:
    if method not in pmb.config.flashers:
        raise RuntimeError(
            f"Flash method {method} is not supported by the"
            " current configuration. However, adding a new"
            " flash method is not that hard, when the flashing"
            " application already exists.\n"
            "Make sure, it is packaged for Alpine Linux, or"
            " package it yourself, and then add it to"
            " pmb/config/__init__.py."
        )
    depends = pmb.config.flashers[method]["depends"]

    # Depends for some flash methods may be different for various pmaports
    # branches, so read them from pmaports.cfg.
    if method == "fastboot":
        pmaports_cfg = pmb.config.pmaports.read_config()
        depends = pmaports_cfg.get("supported_fastboot_depends", "android-tools,avbtool").split(",")
    elif method == "heimdall-bootimg":
        pmaports_cfg = pmb.config.pmaports.read_config()
        depends = pmaports_cfg.get("supported_heimdall_depends", "heimdall,avbtool").split(",")
    elif method == "mtkclient":
        pmaports_cfg = pmb.config.pmaports.read_config()
        depends = pmaports_cfg.get("supported_mtkclient_depends", "mtkclient,android-tools").split(
            ","
        )

    if not isinstance(depends, list):
        raise RuntimeError(f"depends was {type(depends)}, not a list")

    chroot = Chroot.native()

    pmb.chroot.init(chroot)
    pmb.chroot.apk.install(depends, chroot)


def init(device: str, method: str) -> None:
    install_depends(method)

    # Mount folders from host system
    for folder in pmb.config.flash_mount_bind:
        pmb.helpers.mount.bind(folder, Chroot.native() / folder)

    # Mount device chroot inside native chroot (required for kernel/ramdisk)
    mount_device_rootfs(Chroot(ChrootType.ROOTFS, device))
